# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name,missing-function-docstring

import tempfile

import pytest

import SimpleITK as sitk
import numpy as np

from platipy.imaging.tests.data import get_lung_nifti

from platipy.imaging import ImageVisualiser
from platipy.imaging.label.utils import get_com


@pytest.fixture
def nifti_data():

    return get_lung_nifti()


def test_contour_visualisation(nifti_data):

    patient_path = nifti_data.joinpath("LCTSC-Test-S1-201")

    ct_path = next(patient_path.glob("IMAGES/*.nii.gz"))

    structures = {
        struct.name.split(".nii.gz")[0].split("RTSTRUCT_")[-1]: sitk.ReadImage(str(struct))
        for struct in patient_path.glob("STRUCTURES/*.nii.gz")
    }
    img = sitk.ReadImage(str(ct_path))

    vis = ImageVisualiser(img, cut=get_com(structures["HEART"]))
    vis.add_contour(structures)  # Add the contours
    fig = vis.show()

    assert len(fig.axes[0].collections) == 5  # Check correct number of contours

    for idx, structure in enumerate(structures.keys()):
        assert fig.axes[0].legend().get_texts()[idx].get_text() == structure


def test_scalar_overlay_visualisation(nifti_data):

    patient_path = nifti_data.joinpath("LCTSC-Test-S1-201")

    ct_path = next(patient_path.glob("IMAGES/*.nii.gz"))

    structures = {
        struct.name.split(".nii.gz")[0].split("RTSTRUCT_")[-1]: sitk.ReadImage(str(struct))
        for struct in patient_path.glob("STRUCTURES/*.nii.gz")
    }
    img = sitk.ReadImage(str(ct_path))

    vis = ImageVisualiser(img, cut=get_com(structures["HEART"]))
    vis.add_scalar_overlay(structures["HEART"] * 5)  # Add a dummy overlay
    fig = vis.show()

    assert len(fig.axes[0].images) == 2  # Check correct number of images
    img = fig.axes[0].images[-1]
    assert img.get_array().data.sum() == 61295


def test_comparison_overlay_visualisation(nifti_data):

    patient_path = nifti_data.joinpath("LCTSC-Test-S1-201")

    ct_path = next(patient_path.glob("IMAGES/*.nii.gz"))

    structures = {
        struct.name.split(".nii.gz")[0].split("RTSTRUCT_")[-1]: sitk.ReadImage(str(struct))
        for struct in patient_path.glob("STRUCTURES/*.nii.gz")
    }
    img = sitk.ReadImage(str(ct_path))

    vis = ImageVisualiser(img, cut=get_com(structures["HEART"]))
    vis.add_comparison_overlay(structures["HEART"] * 5)  # Add a dummy overlay
    fig = vis.show()

    assert len(fig.axes[0].images) == 1  # Check correct number of images
    img = fig.axes[0].images[0]
    print(img.get_array().data.sum())
    assert np.allclose(img.get_array().data.sum(), 177574, atol=1)



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import click

from platipy.dicom.io.rtstruct_to_nifti import convert_rtstruct


@click.command()
@click.option(
    "--dcm_img",
    "-i",
    required=True,
    help="Directory containing the image series linked to the contour to convert",
)
@click.option(
    "--dcm_rt_file",
    "-r",
    required=True,
    help="Dicom RTStruct file containing the contours to convert",
)
@click.option("--prefix", "-p", default="Struct_", help="Prefix for output files (e.g. Case_01_")
@click.option(
    "--output_dir",
    "-od",
    default=".",
    help="Directory in which to place converted files",
    show_default=True,
)
@click.option("--output_img", "-oi", help="Output name of converted image series")
@click.option(
    "--spacing",
    help="DICOM image spacing override with format x,y,z (0 indicates to leave as is, e.g. 0,0,3)",
)
def click_command(dcm_img, dcm_rt_file, prefix, output_dir, output_img, spacing):
    """
    Convert a DICOM RTSTRUCT file to Nifti format.
    """

    convert_rtstruct(
        dcm_img,
        dcm_rt_file,
        prefix=prefix,
        output_dir=output_dir,
        output_img=output_img,
        spacing=spacing,
    )


if __name__ == "__main__":

    click_command()  # pylint: disable=no-value-for-parameter



# Copyright 2021 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name,missing-function-docstring

import pytest

import SimpleITK as sitk
import numpy as np

from platipy.imaging.tests.data import get_hn_nifti

from platipy.imaging.dose.dvh import calculate_dvh_for_labels, calculate_d_x, calculate_v_x


@pytest.fixture
def nifti_data():

    return get_hn_nifti()


def compute_dvh_for_data(data_path):

    test_pat_path = data_path.joinpath("TCGA_CV_5977")

    ct_image = sitk.ReadImage(
        str(test_pat_path.joinpath("IMAGES/TCGA_CV_5977_1_CT_ONC_NECK_NECK_4.nii.gz"))
    )

    dose = sitk.ReadImage(str(test_pat_path.joinpath("DOSES/TCGA_CV_5977_1_PLAN.nii.gz")))
    dose = sitk.Resample(dose, ct_image)

    structure_names = [
        "BRAINSTEM",
        "MANDIBLE",
        "CTV_60_GY",
        "PTV60",
        "CORD",
        "L_PAROTID",
        "R_PAROTID",
    ]
    structures = {
        s: sitk.ReadImage(
            str(test_pat_path.joinpath("STRUCTURES", f"TCGA_CV_5977_1_RTSTRUCT_{s}.nii.gz"))
        )
        for s in structure_names
    }

    return calculate_dvh_for_labels(dose, structures)


def test_compute_dvh(nifti_data):

    dvh = compute_dvh_for_data(nifti_data)
    assert len(dvh) == 7

    # Check the values for one bin
    assert np.allclose(
        dvh[60.0],
        [
            0.0,
            0.2022032,
            0.9675792,
            0.8746213,
            0.0,
            0.0003158,
            0.0,
        ],
        atol=1e-4,
    )


def test_compute_d_metric(nifti_data):

    dvh = compute_dvh_for_data(nifti_data)
    df_metrics = calculate_d_x(dvh, 95)
    print(list(df_metrics["D95"]))
    assert np.allclose(
        list(df_metrics["D95"]),
        [
            0.035204802259887036,
            36.565625000000026,
            60.25506015037594,
            58.90323404255319,
            6.558703703703705,
            9.182,
            7.33357142857143,
        ],
    )


def test_compute_v_metric(nifti_data):

    dvh = compute_dvh_for_data(nifti_data)
    df_metrics = calculate_v_x(dvh, 40)
    assert np.allclose(
        list(df_metrics["V40"]),
        [
            0.0,
            59.137344,
            190.98043,
            280.18474,
            0.0,
            0.9465217,
            1.3923645,
        ],
        atol=1e-4,
    )



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name

import tempfile
import shutil

import pytest

import SimpleITK as sitk
import numpy as np

from platipy.imaging.tests.data import get_lung_nifti

from platipy.imaging.projects.bronchus.bronchus import (
    generate_lung_mask,
    generate_airway_mask,
)


@pytest.fixture
def bronchus_data():
    """Gets the data needed for the bronchus tests

    Returns:
        pathlib.Path: Path to Data for bronchus test
    """
    return get_lung_nifti()


def assert_lung_mask(lung_mask):
    """Checks that the lung mask looks as expected for the tests in this file"""

    label_shape_statistics_image_filter = sitk.LabelShapeStatisticsImageFilter()
    label_shape_statistics_image_filter.Execute(lung_mask)

    assert np.allclose(label_shape_statistics_image_filter.GetPhysicalSize(1), 2480246, atol=100)

    assert np.allclose(label_shape_statistics_image_filter.GetElongation(1), 1.48, atol=0.01)
    assert np.allclose(label_shape_statistics_image_filter.GetRoundness(1), 0.46, atol=0.01)

    centroid = label_shape_statistics_image_filter.GetCentroid(1)
    assert np.allclose(centroid[0], -1, atol=1)
    assert np.allclose(centroid[1], -169, atol=1)
    assert np.allclose(centroid[2], -476, atol=1)


def assert_bronchus_mask(bronchus_mask):
    """Checks that the bronchus mask looks as expected for the tests in this file"""

    label_shape_statistics_image_filter = sitk.LabelShapeStatisticsImageFilter()
    label_shape_statistics_image_filter.Execute(bronchus_mask)

    print(label_shape_statistics_image_filter.GetPhysicalSize(1))
    print(label_shape_statistics_image_filter.GetElongation(1))
    print(label_shape_statistics_image_filter.GetRoundness(1))
    print(label_shape_statistics_image_filter.GetCentroid(1))
    assert np.allclose(label_shape_statistics_image_filter.GetPhysicalSize(1), 42823, atol=100)

    assert np.allclose(label_shape_statistics_image_filter.GetElongation(1), 1.41, atol=0.01)
    assert np.allclose(label_shape_statistics_image_filter.GetRoundness(1), 0.55, atol=0.01)

    centroid = label_shape_statistics_image_filter.GetCentroid(1)
    assert np.allclose(centroid[0], 8.85, atol=1)
    assert np.allclose(centroid[1], -160, atol=1)
    assert np.allclose(centroid[2], -457, atol=1)


def test_lung_segmentation(bronchus_data):
    """Tests the lung segmentation used as an initial step of bronchus segmentation"""

    patient_path = bronchus_data.joinpath("LCTSC-Test-S1-201")

    ct_path = next(patient_path.glob("IMAGES/*.nii.gz"))
    img = sitk.ReadImage(str(ct_path))

    lung_mask = generate_lung_mask(img)

    assert_lung_mask(lung_mask)


def test_bronchus_segmentation(bronchus_data):
    """Tests the bronchus segmentation algorithm"""

    patient_path = bronchus_data.joinpath("LCTSC-Test-S1-201")

    ct_path = next(patient_path.glob("IMAGES/*.nii.gz"))
    img = sitk.ReadImage(str(ct_path))

    working_dir = tempfile.mkdtemp()

    lung_mask = generate_lung_mask(img)
    bronchus_mask = generate_airway_mask(working_dir, img, lung_mask)

    assert_bronchus_mask(bronchus_mask)

    shutil.rmtree(working_dir)


# def test_bronchus_service(bronchus_data):
#     """An end-to-end test to check that the bronchus service is running as expected"""

#     working_dir = tempfile.mkdtemp()

#     patient_path = bronchus_data.joinpath("LCTSC-Test-S1-201")
#     ct_path = next(patient_path.glob("IMAGES/*.nii.gz"))

#     # Create a data object to be segmented
#     data_object = DataObject()
#     data_object.id = 1
#     data_object.path = str(ct_path)
#     data_object.type = "FILE"

#     # Run the service function
#     result = bronchus_service([data_object], working_dir, BRONCHUS_SETTINGS_DEFAULTS)

#     # Should have returned two output objects
#     assert len(result) == 2

#     lung_mask = sitk.ReadImage(result[0].path)
#     assert_lung_mask(lung_mask)

#     bronchus_mask = sitk.ReadImage(result[1].path)
#     assert_bronchus_mask(bronchus_mask)



# Copyright 2020 University of New South Wales, University of Sydney

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import re
import sys

import pathlib
import pydicom
import numpy as np
import SimpleITK as sitk

from skimage.draw import polygon

from datetime import datetime


logger = logging.getLogger(__name__)


def flatten(itr):
    if type(itr) in (str, bytes, sitk.Image):
        yield itr
    else:
        for x in itr:
            try:
                yield from flatten(x)
            except TypeError:
                yield x


def get_suv_bw_scale_factor(ds):
    # Modified from
    # https://qibawiki.rsna.org/images/6/62/SUV_vendorneutral_pseudocode_happypathonly_20180626_DAC.pdf

    if ds.Units == "CNTS":
        # Try to find the Philips private scale factor")
        return float(ds[0x7053, 0x1000].value)

    assert ds.Modality == "PT"
    assert "DECY" in ds.CorrectedImage
    assert "ATTN" in ds.CorrectedImage
    assert "START" in ds.DecayCorrection
    assert ds.Units == "BQML"

    half_life = float(ds.RadiopharmaceuticalInformationSequence[0].RadionuclideHalfLife)

    if "SeriesTime" in ds:
        series_date_time = ds.SeriesDate + "_" + ds.SeriesTime
    if "." in series_date_time:
        series_date_time = series_date_time[
            : -(len(series_date_time) - series_date_time.index("."))
        ]
    series_date_time = datetime.strptime(series_date_time, "%Y%m%d_%H%M%S")

    if "SeriesTime" in ds:
        start_time = (
            ds.SeriesDate
            + "_"
            + ds.RadiopharmaceuticalInformationSequence[0].RadiopharmaceuticalStartTime
        )
    if "." in start_time:
        start_time = start_time[: -(len(start_time) - start_time.index("."))]
    start_time = datetime.strptime(start_time, "%Y%m%d_%H%M%S")

    decay_time = (series_date_time - start_time).seconds
    injected_dose = float(ds.RadiopharmaceuticalInformationSequence[0].RadionuclideTotalDose)
    decayed_dose = injected_dose * pow(2, -decay_time / half_life)
    patient_weight = float(ds.PatientWeight)
    suv_bw_scale_factor = patient_weight * 1000 / decayed_dose

    return suv_bw_scale_factor


def get_dicom_info_from_description(dicom_object, return_extra=False, sop_class_name="UNKNOWN"):
    """
    Attempts to return some information from a DICOM
    This is typically used for naming converted NIFTI files

    Args:
        dicom_object (pydicom.dataset.FileDataset): The DICOM object
        return_extra (bool, optional): return information that is usually not required

    Returns:
        info (str): Some extracted information
    """
    try:
        dicom_sop_class_name = dicom_object.SOPClassUID.name
    except AttributeError:
        logger.warning("Could not find DICOM SOP Class UID, using %s.", sop_class_name)
        dicom_sop_class_name = sop_class_name

    if "Image" in dicom_sop_class_name:
        # Get the modality
        image_modality = dicom_object.Modality
        logger.info("    Image modality: %s", image_modality)

        if image_modality == "CT":
            # There is typically not much extra information
            # At the moment, we do not return anything for CT imaging
            if return_extra:
                try:
                    protocol_name = dicom_object.ProtocolName

                    if protocol_name != "":
                        return re.sub(r"[^\w]", "_", protocol_name).upper()
                except AttributeError:
                    logger.warning("    Could not find ProtocolName")

            return ""

        elif image_modality == "MR":
            # Not much consistency, but we can get the protocol name
            try:
                protocol_name = re.sub(r"[^\w]", "_", dicom_object.ProtocolName).upper()
            except AttributeError:
                logger.warning("    Could not find ProtocolName")
                protocol_name = ""

            try:
                sequence_name = re.sub(r"[^\w]", "_", dicom_object.SequenceName).upper()
            except AttributeError:
                logger.warning("    Could not find SequenceName")
                sequence_name = ""

            try:
                series_description = re.sub(r"[^\w]", "_", dicom_object.SeriesDescription).upper()
            except AttributeError:
                logger.warning("    Could not find SequenceName")
                series_description = ""

            combined_name = "_".join([protocol_name, sequence_name, series_description])

            while "__" in combined_name:
                combined_name = combined_name.replace("__", "_")

            if protocol_name != "" and not return_extra:
                return protocol_name

            else:
                return combined_name

        elif image_modality == "PT":
            # Not much experience with this
            # We can search through the corrections applied
            # Return whether or not attentuation is applied

            try:
                corrections = dicom_object.CorrectedImage
            except AttributeError:
                corrections = "NONE"

            if "ATTN" in corrections:
                return "AC"
            else:
                return "NAC"


def safe_sort_dicom_image_list(dicom_image_list):
    """
    Sorts a list of DICOM image files based on a DICOM tag value.
    This is a much safer method than reading SliceLocation.
    It takes mandatory DICOM fields (Image Position [Patient]) and (Image Orientation [Patient]).
    The list of DICOM files is sorted by projecting the image position onto the axis normal to the
    place defined by the image orientation.

    This accounts for differences in patient position (e.g. HFS/FFS).

    Args:
        dicom_image_list (list): [description]
    """
    sorted_dict = {}
    for dicom_file in dicom_image_list:
        dcm = pydicom.read_file(dicom_file, force=True)

        image_position = np.array(dcm.ImagePositionPatient, dtype=float)
        image_orientation = np.array(dcm.ImageOrientationPatient, dtype=float)

        image_plane_normal = np.cross(image_orientation[:3], image_orientation[3:])

        slice_location = (image_position * image_plane_normal)[2]

        sorted_dict[dicom_file] = slice_location

    sorter_safe = lambda dcm_file: sorted_dict[dcm_file]

    return sorted(dicom_image_list, key=sorter_safe)


def fix_missing_data(contour_data_list):
    """
    Fixes missing points in contouring using simple linear interpolation


    Args:
        contour_data_list (list): The contour data for each slice

    Returns:
        contour_data (numpy array): Interpolated contour data
    """
    contour_data = np.array(contour_data_list)
    if contour_data.any() == "":
        logger.warning("    Missing values detected.")
        missing_values = np.where(contour_data == "")[0]
        if missing_values.shape[0] > 1:
            logger.warning("    More than one value missing, fixing this isn't implemented yet...")
        else:
            logger.warning("    Only one value missing.")
            missing_index = missing_values[0]
            missing_axis = missing_index % 3
            if missing_axis == 0:
                logger.warning("    Missing value in x axis: interpolating.")
                if missing_index > len(contour_data) - 3:
                    lower_val = contour_data[missing_index - 3]
                    upper_val = contour_data[0]
                elif missing_index == 0:
                    lower_val = contour_data[-3]
                    upper_val = contour_data[3]
                else:
                    lower_val = contour_data[missing_index - 3]
                    upper_val = contour_data[missing_index + 3]
                contour_data[missing_index] = 0.5 * (lower_val + upper_val)
            elif missing_axis == 1:
                logger.warning("    Missing value in y axis: interpolating.")
                if missing_index > len(contour_data) - 2:
                    lower_val = contour_data[missing_index - 3]
                    upper_val = contour_data[1]
                elif missing_index == 0:
                    lower_val = contour_data[-2]
                    upper_val = contour_data[4]
                else:
                    lower_val = contour_data[missing_index - 3]
                    upper_val = contour_data[missing_index + 3]
                contour_data[missing_index] = 0.5 * (lower_val + upper_val)
            else:
                logger.warning("    Missing value in z axis: taking slice value")
                temp = contour_data[2::3].tolist()
                temp.remove("")
                contour_data[missing_index] = np.min(np.array(temp, dtype=np.double))
    return contour_data


def transform_point_set_from_dicom_struct(image, dicom_struct, spacing_override=False):
    """
    This function is used to generate a binary mask from a set of vertices.
    This allows us to convert from DICOM-RTStruct format to any imaging format.

    Args:
        image ([SimpleITK.Image]): The image, used to copy imaging information
            (e.g. resolution, spacing)
        dicom_struct ([pydicom.Dataset]): The DICOM-RTStruct file
        spacing_override (bool | tuple, optional): Overwrite the spacing.
            Set with (axial_spacing, coronal_spacing, sagittal spacing). Defaults to False.

    Returns:
        list, list : final_struct_name_sequence, structure_list
    """
    if spacing_override:
        current_spacing = list(image.GetSpacing())
        new_spacing = tuple(
            [
                current_spacing[k] if spacing_override[k] == 0 else spacing_override[k]
                for k in range(3)
            ]
        )
        image.SetSpacing(new_spacing)

    struct_point_sequence = dicom_struct.ROIContourSequence
    struct_name_sequence = [
        "_".join(i.ROIName.split()) for i in dicom_struct.StructureSetROISequence
    ]

    structure_list = []
    final_struct_name_sequence = []

    for structIndex, structure_name in enumerate(struct_name_sequence):
        image_blank = np.zeros(image.GetSize()[::-1], dtype=np.uint8)
        logger.info("    Converting structure %d with name: %s", structIndex, structure_name)

        if structIndex >= len(struct_point_sequence):
            logger.warning("    Contour sequence is missing, skipping.")
            continue

        if not hasattr(struct_point_sequence[structIndex], "ContourSequence"):
            logger.warning("    No contour sequence found for this structure, skipping.")
            continue

        if len(struct_point_sequence[structIndex].ContourSequence) == 0:
            logger.warning("    Contour sequence is empty, skipping.")
            continue

        if (
            not struct_point_sequence[structIndex].ContourSequence[0].ContourGeometricType
            == "CLOSED_PLANAR"
        ):
            logger.warning("    This is not a closed planar structure, skipping.")
            continue

        for sl in range(len(struct_point_sequence[structIndex].ContourSequence)):

            contour_data = fix_missing_data(
                struct_point_sequence[structIndex].ContourSequence[sl].ContourData
            )

            struct_slice_contour_data = np.array(contour_data, dtype=np.double)
            vertexArr_physical = struct_slice_contour_data.reshape(
                struct_slice_contour_data.shape[0] // 3, 3
            )

            point_arr = np.array(
                [image.TransformPhysicalPointToIndex(i) for i in vertexArr_physical]
            ).T

            [xVertexArr_image, yVertexArr_image] = point_arr[[0, 1]]
            zIndex = point_arr[2][0]

            if np.any(point_arr[2] != zIndex):
                logger.error("    Axial slice index varies in contour. Quitting now.")
                logger.error("    Structure:   %s", structure_name)
                logger.error("    Slice index: %s", zIndex)
                quit()

            if zIndex >= image.GetSize()[2]:
                logger.warning("    Slice index greater than image size. Skipping slice.")
                logger.warning("    Structure:   %s", structure_name)
                logger.warning("    Slice index: %s", zIndex)
                continue

            sliceArr = np.zeros(image.GetSize()[:2], dtype=np.uint8)
            filledIndicesX, filledIndicesY = polygon(
                xVertexArr_image, yVertexArr_image, shape=sliceArr.shape
            )
            sliceArr[filledIndicesX, filledIndicesY] = 1
            image_blank[zIndex] += sliceArr.T

        struct_image = sitk.GetImageFromArray(1 * (image_blank > 0))
        struct_image.CopyInformation(image)
        structure_list.append(sitk.Cast(struct_image, sitk.sitkUInt8))
        structure_name_clean = re.sub(r"[^\w]", "_", structure_name).upper()
        while "__" in structure_name_clean:
            structure_name_clean = structure_name_clean.replace("__", "_")
        final_struct_name_sequence.append(structure_name_clean)

    return final_struct_name_sequence, structure_list


def process_dicom_file_list(dicom_file_list, parent_sorting_field="PatientName", verbose=False):

    """
    Organise the DICOM files by the series UID
    """
    dicom_series_dict_parent = {}

    for i, dicom_file in enumerate(sorted(dicom_file_list)):
        if verbose is True:
            logger.debug("  Sorting file %d", i)

        dicom_file = dicom_file.as_posix()

        if "dicomdir" in dicom_file.lower():
            logger.warning(
                "DICOMDIR is not supported in this tool, images are read directly. Skipping."
            )
            continue

        dicom_object = pydicom.read_file(dicom_file, force=True)

        parent_sorting_field_data = dicom_object[parent_sorting_field].value

        if parent_sorting_field_data not in dicom_series_dict_parent.keys():
            dicom_series_dict_parent[parent_sorting_field_data] = {}

        series_uid = dicom_object.SeriesInstanceUID

        if series_uid not in dicom_series_dict_parent[parent_sorting_field_data].keys():
            dicom_series_dict_parent[parent_sorting_field_data][series_uid] = [dicom_file]

        else:
            dicom_series_dict_parent[parent_sorting_field_data][series_uid].append(dicom_file)

    return dicom_series_dict_parent


def process_dicom_series(
    dicom_series_dict,
    series_uid,
    parent_sorting_field="PatientName",
    return_extra=True,
    individual_file=False,
    initial_sop_class_name_default="UNKNOWN",
):
    if not individual_file:
        logger.info("  Processing series UID: %s", series_uid)
        dicom_file_list = dicom_series_dict[series_uid]
    else:
        logger.info("  Processing individual file: %s", individual_file)
        dicom_file_list = [individual_file]

    logger.info("  Number of DICOM files: %d", len(dicom_file_list))

    initial_dicom = pydicom.read_file(dicom_file_list[0])

    # Get the data in the parent sorting field, clean with RegEx
    parent_sorting_data = re.sub(
        r"[^\w]", "_", str(initial_dicom[parent_sorting_field].value)
    ).upper()

    if parent_sorting_data == "":
        logger.error(
            "Could not find any data in %s. This is very bad, the data cannot be sorted properly.",
            parent_sorting_field,
        )
        """
        ! TO DO
        Implement a routine to let a user correlate a root directory with a name
        """
        parent_sorting_data = "TEMP"

    try:
        initial_dicom_sop_class_name = initial_dicom.SOPClassUID.name
    except AttributeError:
        logger.warning(
            "Could not find DICOM SOP Class UID, using %s.", initial_sop_class_name_default
        )
        initial_dicom_sop_class_name = initial_sop_class_name_default

    try:
        study_uid = initial_dicom.StudyInstanceUID
    except AttributeError:
        study_uid = "00001"

    """
    ! TO DO
    Need to check for secondary capture image storage
    This can include JPEGs with written information on them
    This is typically not very useful
    We can dump it to file
    Or just save the DICOM file in the folder of interest

    Not a big problem, sort out another day
    """

    # Check the potential types of DICOM files
    if (
        "Image" in initial_dicom_sop_class_name
        and initial_dicom_sop_class_name != "Secondary Capture Image Storage"
    ):
        # Load as an primary image

        sorted_file_list = safe_sort_dicom_image_list(dicom_file_list)

        try:
            image = sitk.ReadImage(sorted_file_list)
        except RuntimeError:
            logger.warning("  Could not read image into SimpleITK.")
            logger.info("  Processing files individually.")

            for dicom_file in dicom_file_list:
                return process_dicom_series(
                    dicom_series_dict,
                    series_uid,
                    parent_sorting_field=parent_sorting_field,
                    return_extra=return_extra,
                    individual_file=dicom_file,
                    initial_sop_class_name_default=initial_sop_class_name_default,
                )

        dicom_file_metadata = {
            "parent_sorting_data": parent_sorting_data,
            "study_uid": study_uid,
        }

        """
        ! TO DO - integrity check
            Read in all the files here, check the slice location and determine if any are missing
        """
        if initial_dicom.Modality == "PT":

            # scaling_factor = get_suv_bw_scale_factor(initial_dicom)
            # image *= scaling_factor

            # !TO DO
            # Work on PET SUV conversion
            pass

        """
        ! CHECKPOINT
        Some DCE MRI sequences have the same series UID
        Here we check the sequence name, and split if necessary
        """

        if initial_dicom.Modality == "MR":

            try:
                sequence_names = np.unique(
                    [pydicom.read_file(x).SequenceName for x in dicom_file_list]
                )

                sequence_dict = {}
                for dcm_name in dicom_file_list:
                    dcm_obj = pydicom.read_file(dcm_name)
                    var = dcm_obj.SequenceName
                    if var not in sequence_dict.keys():
                        sequence_dict[var] = [dcm_name]
                    else:
                        sequence_dict[var].append(dcm_name)

            except AttributeError:
                try:
                    logger.warning(
                        "    MRI sequence name not found. The SeriesDescription will be used instead."
                    )

                    sequence_names = np.unique(
                        [pydicom.read_file(x).SeriesDescription for x in dicom_file_list]
                    )

                    sequence_dict = {}
                    for dcm_name in dicom_file_list:
                        dcm_obj = pydicom.read_file(dcm_name)
                        var = dcm_obj.SeriesDescription
                        if var not in sequence_dict.keys():
                            sequence_dict[var] = [dcm_name]
                        else:
                            sequence_dict[var].append(dcm_name)

                except AttributeError:
                    logger.warning(
                        "    MRI SeriesDescription not found. The AcquisitionComments will be used instead."
                    )

                    sequence_names = np.unique(
                        [pydicom.read_file(x).AcquisitionComments for x in dicom_file_list]
                    )

                    sequence_dict = {}
                    for dcm_name in dicom_file_list:
                        dcm_obj = pydicom.read_file(dcm_name)
                        var = dcm_obj.AcquisitionComments
                        if var not in sequence_dict.keys():
                            sequence_dict[var] = [dcm_name]
                        else:
                            sequence_dict[var].append(dcm_name)

            if initial_dicom.Manufacturer == "GE MEDICAL SYSTEMS":
                # GE use the DICOM tag (0019, 10a2) [Raw data run number]
                # in Diffusion weighted MRI sequences
                # We need to separate this out to get the difference sequences

                if initial_dicom.SeriesDescription == "Diffusion Weighted":

                    # num_sequences = int( (initial_dicom[(0x0025, 0x1007)]) / (initial_dicom[(0x0021, 0x104f)]) )
                    # number_of_images / images_per_seq
                    num_images_per_seq = initial_dicom[(0x0021, 0x104F)].value

                    sequence_names = np.unique(
                        [
                            f"DWI_{str( ( pydicom.read_file(x)['InstanceNumber'].value - 1) // num_images_per_seq )}"
                            for x in dicom_file_list
                        ]
                    )

                    sequence_name_index_dict = {
                        name: index for index, name in enumerate(sequence_names)
                    }

                    sequence_dict = {}
                    for dcm_name in dicom_file_list:
                        dcm_obj = pydicom.read_file(dcm_name)
                        var = f"DWI_{str( ( dcm_obj['InstanceNumber'].value - 1) // num_images_per_seq )}"
                        var_to_index = sequence_name_index_dict[var]

                        if var_to_index not in sequence_dict.keys():
                            sequence_dict[var_to_index] = [dcm_name]
                        else:
                            sequence_dict[var_to_index].append(dcm_name)

                    sequence_names = sorted(sequence_dict.keys())

            if len(sequence_names) > 1:
                logger.warning("  Two MR sequences were found under a single series UID.")
                logger.warning("  These will be split into separate images.")

                # Split up the DICOM file list by sequence name
                for sequence_name in sequence_names:

                    dicom_file_list_by_sequence = sequence_dict[sequence_name]

                    logger.info(sequence_name)
                    logger.info(len(dicom_file_list_by_sequence))

                    sorted_file_list = safe_sort_dicom_image_list(dicom_file_list_by_sequence)

                    initial_dicom = pydicom.read_file(sorted_file_list[0], force=True)

                    image_by_sequence = sitk.ReadImage(sorted_file_list)

                    dicom_file_metadata_by_sequence = {
                        "parent_sorting_data": parent_sorting_data,
                        "study_uid": study_uid,
                    }

                    yield "IMAGES", dicom_file_metadata_by_sequence, initial_dicom, image_by_sequence
                return  # Stop iteration

        yield "IMAGES", dicom_file_metadata, initial_dicom, image

    if "Structure" in initial_dicom_sop_class_name:
        # Load as an RT structure set
        # This should be done individually for each file

        logger.info("      Number of files: %d", len(dicom_file_list))
        for index, dicom_file in enumerate(dicom_file_list):
            dicom_object = pydicom.read_file(dicom_file, force=True)

            # We must also read in the corresponding DICOM image
            # This can be found by matching the references series UID to the series UID

            """
            ! TO DO
            What happens if there is an RT structure set with different referenced sequences?
            """

            # Get the "ReferencedFrameOfReferenceSequence", first item
            referenced_frame_of_reference_item = dicom_object.ReferencedFrameOfReferenceSequence[0]

            # Get the "RTReferencedStudySequence", first item
            # This retrieves the study UID
            # This might be useful, but would typically match the actual StudyInstanceUID in the
            # DICOM object
            rt_referenced_series_item = (
                referenced_frame_of_reference_item.RTReferencedStudySequence[0]
            )

            # Get the "RTReferencedSeriesSequence", first item
            # This retreives the actual referenced series UID, which we need to match imaging
            # parameters
            rt_referenced_series_again_item = rt_referenced_series_item.RTReferencedSeriesSequence[
                0
            ]

            # Get the appropriate series instance UID
            image_series_uid = rt_referenced_series_again_item.SeriesInstanceUID
            logger.info("      Item %s: Matched SeriesInstanceUID = %s", index, image_series_uid)

            # Read in the corresponding image
            sorted_file_list = safe_sort_dicom_image_list(dicom_series_dict[image_series_uid])
            image = sitk.ReadImage(sorted_file_list)

            initial_dicom = pydicom.read_file(sorted_file_list[0], force=True)

            (
                structure_name_list,
                structure_image_list,
            ) = transform_point_set_from_dicom_struct(image, dicom_object)

            dicom_file_metadata = {
                "parent_sorting_data": parent_sorting_data,
                "study_uid": study_uid,
                "structure_name_list": structure_name_list,
            }

            yield "STRUCTURES", dicom_file_metadata, dicom_object, structure_image_list

    if "Dose" in initial_dicom_sop_class_name:
        # Load as an RT Dose distribution
        # This should be done individually for each file

        logger.info("      Number of files: %d", len(dicom_file_list))
        for index, dicom_file in enumerate(dicom_file_list):
            dicom_object = pydicom.read_file(dicom_file, force=True)

            """
            ! CHECKPOINT
            There should only be a single RT dose file (with each series UID)
            If there are more, yield each
            """

            initial_dicom = pydicom.read_file(dicom_file, force=True)

            dicom_file_metadata = {
                "parent_sorting_data": parent_sorting_data,
                "study_uid": study_uid,
            }

            # We must read in as a float otherwise when we multiply by one later it will not work!
            raw_dose_image = sitk.ReadImage(dicom_file, sitk.sitkFloat32)

            dose_grid_scaling = dicom_object.DoseGridScaling

            logger.debug("  Dose grid scaling: %.5f Gy", dose_grid_scaling)

            scaled_dose_image = raw_dose_image * dose_grid_scaling

            yield "DOSES", dicom_file_metadata, dicom_object, scaled_dose_image

        """
        ! TO DO
        1. (DONE) Implement conversion of dose files (to NIFTI images)
        2. Implement conversion of RT plan files to text dump
        3. Do something with other files (e.g. Deformable Image Registration stuff)
        """

    return


def write_output_data_to_disk(
    output_data_dict,
    output_directory="./",
    output_file_suffix=".nii.gz",
    overwrite_existing_files=False,
):
    """
    Write output to disk
    """
    if output_data_dict is None:
        return

    filename_fields = [i for i in output_data_dict.keys() if i != "parent_sorting_data"]
    parent_sorting_data = output_data_dict["parent_sorting_data"]

    files_written = {}

    """
    Write the the converted images to disk

    ! CONSIDER
    We could simply write as we go?
    Pro: save memory, important if processing very large files
    Con: Reading as we go allows proper indexing

    """

    for field in filename_fields:
        logger.info("  Writing files for field: %s", field)
        p = pathlib.Path(output_directory) / parent_sorting_data / field
        p.mkdir(parents=True, exist_ok=True)
        files_written[field] = []

        for field_filename_base, field_list in output_data_dict[field].items():
            # Check if there is a list of images with matching names
            # This will depend on the name format chosen
            # If there is a list, we append an index as we write to disk

            if isinstance(field_list, (tuple, list)):
                # Flatten
                field_list_flat = list(flatten(field_list))

                # Iterate
                for suffix, file_to_write in enumerate(field_list_flat):
                    field_filename = field_filename_base + f"_{suffix}"

                    # Some cleaning
                    while "__" in field_filename:
                        field_filename = field_filename.replace("__", "_")

                    while field_filename[-1] == "_":
                        field_filename = field_filename[:-1]

                    # Save image!
                    output_name = (
                        pathlib.Path(output_directory)
                        / parent_sorting_data
                        / field
                        / (field_filename + output_file_suffix)
                    )
                    files_written[field].append(output_name)

                    if output_name.is_file():
                        logger.warning("  File exists: %s", output_name)

                        if overwrite_existing_files:
                            logger.warning("  You have selected to overwrite existing files.")

                        else:
                            logger.info(
                                "  You have selected to NOT overwrite existing files. Continuing."
                            )
                            continue

                    sitk.WriteImage(file_to_write, output_name.as_posix())

            else:
                field_filename = field_filename_base
                file_to_write = field_list

                # Some cleaning
                while "__" in field_filename:
                    field_filename = field_filename.replace("__", "_")

                while field_filename[-1] == "_":
                    field_filename = field_filename[:-1]

                # Save image!
                """
                ! TO DO
                Use pathlib, and perform some checks so we don"t overwrite anything!
                """
                output_name = (
                    pathlib.Path(output_directory)
                    / parent_sorting_data
                    / field
                    / (field_filename + output_file_suffix)
                )
                files_written[field].append(output_name)

                if output_name.is_file():
                    logger.warning("  File exists: %s", output_name)

                    if overwrite_existing_files:
                        logger.warning("  You have selected to overwrite existing files.")

                    else:
                        logger.info(
                            "  You have selected to NOT overwrite existing files. Continuing."
                        )
                        continue

                sitk.WriteImage(file_to_write, output_name.as_posix())

    return files_written


def process_dicom_directory(
    dicom_directory,
    parent_sorting_field="PatientName",
    output_image_name_format="{parent_sorting_data}_{study_uid_index}_{Modality}_{image_desc}_{SeriesNumber}",
    output_structure_name_format="{parent_sorting_data}_{study_uid_index}_{Modality}_{structure_name}",
    output_dose_name_format="{parent_sorting_data}_{study_uid_index}_{DoseSummationType}",
    return_extra=True,
    output_directory="./",
    output_file_suffix=".nii.gz",
    overwrite_existing_files=False,
    write_to_disk=True,
    verbose=False,
    initial_sop_class_name_default="UNKNOWN",
):

    # Check dicom_directory type
    if isinstance(dicom_directory, str) or isinstance(dicom_directory, pathlib.Path):
        # Get all the DICOM files in the given directory
        root_path = pathlib.Path(dicom_directory)
        # Find files ending with .dcm, .dc3
        dicom_file_list = [
            p
            for p in root_path.glob("**/*")
            if p.name.lower().endswith(".dcm") or p.name.lower().endswith(".dc3")
        ]

    elif hasattr(dicom_directory, "__iter__"):
        dicom_file_list = []
        for dicom_dir in dicom_directory:
            # Get all the DICOM files in each directory
            root_path = pathlib.Path(dicom_dir)
            # Find files ending with .dcm, .dc3
            dicom_file_list += [
                p
                for p in root_path.glob("**/*")
                if p.name.lower().endswith(".dcm") or p.name.lower().endswith(".dc3")
            ]

    if len(dicom_file_list) == 0:
        logger.info("No DICOM files found in input directory. Exiting now.")
        return

    # Process the DICOM files
    # This returns a dictionary (of dictionaries):
    #   {parent_data (e.g. PatientName): {series_UID_1: [list_of_DICOM_files],
    #                                    {series_UID_2: [list_of_DICOM_files], ...
    #   parent_data_2                  : {series_UID_1: [list_of_DICOM_files],
    #                                    {series_UID_2: [list_of_DICOM_files], ...
    #   ...     }
    dicom_series_dict_parent = process_dicom_file_list(
        dicom_file_list, parent_sorting_field=parent_sorting_field, verbose=verbose
    )

    if dicom_series_dict_parent is None:
        logger.info("No valid DICOM files found. Ending.")
        return None

    output = {}

    for parent_data, dicom_series_dict in dicom_series_dict_parent.items():
        logger.info("Processing data for %s = %s.", parent_sorting_field, parent_data)
        logger.info("  Number of DICOM series = %d", len(dicom_series_dict.keys()))

        # Set up the output data
        # This stores the SimpleITK images and file names
        output_data_dict = {}

        # Set up the study UID dict
        # This helps match structure sets to relevant images
        # And paired images to each other (e.g. PET/CT)
        study_uid_dict = {}

        # Give some user feedback
        logger.debug("  Output image name format: %s", output_image_name_format)
        logger.debug("  Output structure name format: %s", output_structure_name_format)
        logger.debug("  Output dose name format: %s", output_dose_name_format)

        # For each unique series UID, process the DICOM files
        for series_uid in dicom_series_dict.keys():

            # This function returns four values
            # 1. dicom_type: This is IMAGES, STRUCTURES, DOSES, etc
            # 2. dicom_file_metadata: Some special metadata extracted from the DICOM header
            # 3. initial_dicom: The first DICOM in the series. For doses and structures there is
            #    (usually) only one DICOM anyway
            # 4. dicom_file_data: The actual SimpleITK image data

            for (
                dicom_type,
                dicom_file_metadata,
                initial_dicom,
                dicom_file_data,
            ) in process_dicom_series(
                dicom_series_dict=dicom_series_dict,
                series_uid=series_uid,
                parent_sorting_field=parent_sorting_field,
                return_extra=return_extra,
                initial_sop_class_name_default=initial_sop_class_name_default,
            ):

                # Step 1
                # Check the parent sorting field is consistent
                # This would usually be the PatientName, PatientID, or similar
                # Occasionally these will both be blank

                parent_sorting_data = dicom_file_metadata["parent_sorting_data"]

                if "parent_sorting_data" not in output_data_dict.keys():
                    output_data_dict["parent_sorting_data"] = parent_sorting_data

                else:
                    if parent_sorting_data != output_data_dict["parent_sorting_data"]:
                        logger.error(
                            "A conflict was found for the parent sorting field " "(%s): %s}",
                            parent_sorting_field,
                            parent_sorting_data,
                        )
                        logger.error("Quitting now.")
                        print(dicom_series_dict_parent.keys())
                        sys.exit()
                    else:
                        logger.info(
                            "  Parent sorting field (%s) match found: %s",
                            parent_sorting_field,
                            parent_sorting_data,
                        )

                # Step 2
                # Get the study UID
                # Used for indexing DICOM series

                study_uid = dicom_file_metadata["study_uid"]

                if study_uid not in study_uid_dict.keys():
                    try:
                        study_uid_index = max(study_uid_dict.values()) + 1
                    except AttributeError:
                        study_uid_index = 0  # Study UID dict might not exist
                    except ValueError:
                        study_uid_index = 0  # Study UID dict might be empty

                    logger.info("  Setting study instance UID index: %s", study_uid_index)

                    study_uid_dict[study_uid] = study_uid_index

                else:
                    logger.info(
                        "  Study instance UID index already exists: %s", study_uid_dict[study_uid]
                    )

                # Step 3
                # Generate names for output files

                # Special names
                # ! This can be defined once at the start of the function
                special_name_fields = [
                    "parent_sorting_data",
                    "study_uid_index",
                    "image_desc",
                    "structure_name",
                ]

                # Get the image description (other special names are already defined above)
                image_desc = get_dicom_info_from_description(
                    initial_dicom, return_extra=return_extra
                )

                # Get all the fields from the user-given name format
                if dicom_type == "IMAGES":
                    all_naming_fields = [
                        i[i.find("{") + 1 :]
                        for i in output_image_name_format.split("}")
                        if len(i) > 0
                    ]
                elif dicom_type == "STRUCTURES":
                    all_naming_fields = [
                        i[i.find("{") + 1 :]
                        for i in output_structure_name_format.split("}")
                        if len(i) > 0
                    ]
                elif dicom_type == "DOSES":
                    all_naming_fields = [
                        i[i.find("{") + 1 :]
                        for i in output_dose_name_format.split("}")
                        if len(i) > 0
                    ]

                # Now exclude those that aren't derived from the DICOM header
                dicom_header_tags = [i for i in all_naming_fields if i not in special_name_fields]

                naming_info_dict = {}
                for dicom_field in dicom_header_tags:
                    try:
                        dicom_field_value = initial_dicom[dicom_field].value
                    except (AttributeError, KeyError):
                        logger.warning(
                            "  Could not find DICOM header %s. Setting as 0 to "
                            "preserve naming convention.",
                            dicom_field,
                        )
                        dicom_field_value = 0
                    naming_info_dict[dicom_field] = dicom_field_value

                if dicom_type == "IMAGES":

                    output_name = output_image_name_format.format(
                        parent_sorting_data=parent_sorting_data,
                        study_uid_index=study_uid_dict[study_uid],
                        image_desc=image_desc,
                        **naming_info_dict,
                    )

                    if "IMAGES" not in output_data_dict.keys():
                        # Make a new entry
                        output_data_dict["IMAGES"] = {output_name: dicom_file_data}

                    else:
                        # First check if there is another image of the same name

                        if output_name not in output_data_dict["IMAGES"].keys():
                            output_data_dict["IMAGES"][output_name] = dicom_file_data

                        else:
                            logger.info("      An image with this name exists, appending.")

                            if hasattr(output_data_dict["IMAGES"][output_name], "__iter__"):
                                output_data_dict["IMAGES"][output_name] = list(
                                    [output_data_dict["IMAGES"][output_name]]
                                )

                            output_data_dict["IMAGES"][output_name].append(dicom_file_data)

                elif dicom_type == "STRUCTURES":

                    for structure_name, structure_image in zip(
                        dicom_file_metadata["structure_name_list"], dicom_file_data
                    ):

                        output_name = output_structure_name_format.format(
                            parent_sorting_data=parent_sorting_data,
                            study_uid_index=study_uid_dict[study_uid],
                            image_desc=image_desc,
                            structure_name=structure_name,
                            **naming_info_dict,
                        )

                        if "STRUCTURES" not in output_data_dict.keys():
                            # Make a new entry
                            output_data_dict["STRUCTURES"] = {output_name: structure_image}

                        else:
                            # First check if there is another structure of the same name

                            if output_name not in output_data_dict["STRUCTURES"].keys():
                                output_data_dict["STRUCTURES"][output_name] = structure_image

                            else:
                                logger.info("      A structure with this name exists, appending.")
                                if hasattr(
                                    output_data_dict["STRUCTURES"][output_name], "__iter__"
                                ):
                                    output_data_dict["STRUCTURES"][output_name] = list(
                                        [output_data_dict["STRUCTURES"][output_name]]
                                    )

                                output_data_dict["STRUCTURES"][output_name].append(structure_image)

                elif dicom_type == "DOSES":

                    output_name = output_dose_name_format.format(
                        parent_sorting_data=parent_sorting_data,
                        study_uid_index=study_uid_dict[study_uid],
                        **naming_info_dict,
                    )

                    if "DOSES" not in output_data_dict.keys():
                        # Make a new entry
                        output_data_dict["DOSES"] = {output_name: dicom_file_data}

                    else:
                        # First check if there is another image of the same name

                        if output_name not in output_data_dict["DOSES"].keys():
                            output_data_dict["DOSES"][output_name] = dicom_file_data

                        else:
                            logger.info("      An image with this name exists, appending.")

                            if isinstance(output_data_dict["DOSES"][output_name], sitk.Image):
                                output_data_dict["DOSES"][output_name] = list(
                                    [output_data_dict["DOSES"][output_name]]
                                )

                            output_data_dict["DOSES"][output_name].append(dicom_file_data)

        if write_to_disk:
            output[str(parent_data)] = write_output_data_to_disk(
                output_data_dict=output_data_dict,
                output_directory=output_directory,
                output_file_suffix=output_file_suffix,
                overwrite_existing_files=overwrite_existing_files,
            )
        else:
            output[str(parent_data)] = output_data_dict

    """
    TO DO!
    Memory issue with output_data_dict
    Use in inner loop, reset output_data_dict
    """

    return output



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import matplotlib.pyplot as plt
import numpy as np

from platipy.imaging.label.fusion import process_probability_image
from platipy.imaging.utils.crop import label_to_roi, crop_to_roi


def gen_primes():
    """Generate an infinite sequence of prime numbers."""
    # Maps composites to primes witnessing their compositeness.
    # This is memory efficient, as the sieve is not "run forward"
    # indefinitely, but only as long as required by the current
    # number being tested.
    #
    D = {}

    # The running integer that's checked for primeness
    q = 2

    while True:
        if q not in D:
            # q is a new prime.
            # Yield it and mark its first multiple that isn't
            # already marked in previous iterations
            #
            yield q
            D[q * q] = [q]
        else:
            # q is composite. D[q] is the list of primes that
            # divide it. Since we've reached q, we no longer
            # need it in the map, but we'll mark the next
            # multiples of its witnesses to prepare for larger
            # numbers
            #
            for p in D[q]:
                D.setdefault(p + q, []).append(p)
            del D[q]

        q += 1


def quick_optimise_probability(
    metric_function,
    manual_contour,
    probability_image,
    p_0=0.5,
    delta=0.5,
    tolerance=0.01,
    mode="min",
    create_figure=False,
    auto_crop=True,
    metric_args={},
):
    """Optimise the probability threshold used to generate a binary segmentation.
    This is a simple parameter sweep, with linearly decreasing resolution. It will usually converge
    between 5 and 10 iterations.

    Args:
        metric_function (function to return float): The metric function, this takes in two binary
            masks (a reference, and test [SimpleITK.Image]) and returns a metric (as a float).
            Typical choices would be the DSC, a surface distance metric, dose difference,
            relative volume. Additional arguments can be passed in through `metric_args` argument.
        manual_contour (SimpleITK.Image): The reference (manual) contour.
        probability_image (SimpleITK.Image): The probability map from which the optimal threshold
            will be derived. This does NOT have to be scaled to [0,1].
        p_0 (float, optional): Initial guess of the optimal threshold. Defaults to 0.5.
        delta (float, optional): The window size of the optimiser. Defaults to 0.5.
        tolerance (float, optional): If the metric changes by an amount less that `tolerance`
            the optimiser will stop. Defaults to 0.01.
        mode (str, optional): Specifies whether the metric should be maximised ("max") or minimised
            ("min"). Defaults to "min".
        create_figure (bool, optional): Create a matplotlib figure showing the optimisation. This
            is not returned, so make sure you can capture this (e.g. using IPython).
            Defaults to False.
        auto_crop (bool, optional): Crop the image volumes to the region of interest. Speeds up the
            process signficiantly so don't turn off unless you have a good reason to!
            Defaults to True.
        metric_args (dict, optional): Additional arguments passes to the metric function. This
            could be useful if you are calculating a dose-based metric and require a dose grid to
            be passed to the metric function. Defaults to {}.

    Returns:
        tuple (float, float): The optimal probability, optimal metric value.
    """
    # Auto crop images
    if auto_crop:
        cb_size, cb_index = label_to_roi(
            (manual_contour > 0) | (probability_image > 0), expansion_mm=10
        )
        manual_contour = crop_to_roi(manual_contour, cb_size, cb_index)
        probability_image = crop_to_roi(probability_image, cb_size, cb_index)

    # Set up
    n_iter = 0
    p_best = p_0

    auto_contour = process_probability_image(probability_image, threshold=p_0)

    m_n = metric_function(manual_contour, auto_contour, **metric_args)
    m_best = m_n

    print("Starting optimisation.")
    print(f"n = 0 | p = {p_best:.3f} | metric = {m_n:.3f}")

    p_list = [p_best]
    m_list = [m_best]

    improv = 0

    while np.abs(improv) > tolerance or n_iter <= 3:

        n_iter += 1
        m_n = m_best

        p_new = [
            p_best - 3 * delta / 4,
            p_best - delta / 2,
            p_best - delta / 4,
            p_best + delta / 4,
            p_best + delta / 2,
            p_best + 3 * delta / 4,
        ]
        m_new = [
            metric_function(
                manual_contour,
                process_probability_image(probability_image, threshold=p),
                **metric_args,
            )
            for p in p_new
        ]

        p_list = p_list + p_new
        m_list = m_list + m_new

        if mode == "min":
            p_best = p_list[np.argmin(m_list)]
            m_best = np.min(m_list)
        elif mode == "max":
            p_best = p_list[np.argmax(m_list)]
            m_best = np.max(m_list)

        improv = m_best - m_n

        delta /= 4

        print(f"n = {n_iter} | p = {p_best:.3f} | metric = {m_best:.3f}")

    if create_figure:
        fig, ax = plt.subplots(1, 1)
        ax.scatter(p_list, m_list, c="k", zorder=1)
        ax.plot(*list(zip(*sorted(zip(p_list, m_list)))), c="k", zorder=1)
        ax.scatter(
            (p_best), (m_best), c="r", label=f"Optimum ({p_best:.2f},{m_best:.2f})", zorder=2
        )
        ax.set_xlim(0, 1)
        ax.set_xlabel("Probability Difference (from Optimal)")
        ax.set_ylabel("Metric Value")
        ax.grid()
        ax.set_axisbelow(True)
        ax.set_title(f"Optimiser | {metric_function.__name__}, mode = {mode}")
        ax.legend()
        fig.show()

    return p_best, m_best



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import SimpleITK as sitk
import numpy as np

from scipy.ndimage.measurements import center_of_mass

from platipy.imaging.utils.math import gen_primes


def correct_volume_overlap(binary_label_dict, assign_overlap_to_largest=True):
    """
    Label structures by primes
    Smallest prime = largest volume
    """

    # Calculate volume
    f_vol = lambda x: sitk.GetArrayViewFromImage(x).sum()
    volume_dict = {i: f_vol(binary_label_dict[i]) for i in binary_label_dict.keys()}

    keys, vals = zip(*volume_dict.items())
    if assign_overlap_to_largest:
        volume_rank = np.argsort(vals)[::-1]
    else:
        volume_rank = np.argsort(vals)

    # print(keys, volume_rank)

    ranked_names = np.array(keys)[volume_rank]

    # Get overlap (this is used to reconstruct labels)
    combined_label = sum(binary_label_dict.values()) > 0

    # Prime encode each binary label
    prime_labelled_image = prime_encode_structure_list(
        [binary_label_dict[i] for i in ranked_names]
    )

    # Remove overlap (by assigning to binary volume)
    output_label_dict = {}
    for p, label in zip(gen_primes(), ranked_names):
        output_label_dict[label] = combined_label * (sitk.Modulus(prime_labelled_image, p) == 0)

        combined_label = sitk.MaskNegated(combined_label, output_label_dict[label])

    return output_label_dict


def get_com(label, as_int=True, real_coords=False):
    """
        Get centre of mass of a SimpleITK.Image

    Args:
        label (sitk.Image): Label mask image.
        as_int (bool, optional): Returns each components as int if true. Defaults to True.
        real_coords (bool, optional): Return coordinates in physical space if true. Defaults to
            False.

    Returns:
        list: List of coordinates
    """
    arr = sitk.GetArrayFromImage(label)
    com = center_of_mass(arr)

    if real_coords:
        com = label.TransformContinuousIndexToPhysicalPoint(com[::-1])

    else:
        if as_int:
            com = [int(i) for i in com]

    return com


def vectorised_transform_index_to_physical_point(image, point_array, rotate=True):
    """
    Transforms a set of points from array indices to real-space
    """
    if rotate:
        spacing = image.GetSpacing()[::-1]
        origin = image.GetOrigin()[::-1]
    else:
        spacing = image.GetSpacing()
        origin = image.GetOrigin()
    return point_array * spacing + origin


def vectorised_transform_physical_point_to_index(image, point_array, rotate=True):
    """
    Transforms a set of points from real-space to array indices
    """
    if rotate:
        spacing = image.GetSpacing()[::-1]
        origin = image.GetOrigin()[::-1]
    else:
        spacing = image.GetSpacing()
        origin = image.GetOrigin()
    return (point_array - origin) / spacing


def generate_primes():
    """Generate an infinite sequence of prime numbers."""
    # Maps composites to primes witnessing their compositeness.
    # This is memory efficient, as the sieve is not "run forward"
    # indefinitely, but only as long as required by the current
    # number being tested.
    #
    d = {}

    # The running integer that's checked for primeness
    q = 2

    while True:
        if q not in d:
            # q is a new prime.
            # Yield it and mark its first multiple that isn't
            # already marked in previous iterations
            #
            yield q
            d[q * q] = [q]
        else:
            # q is composite. D[q] is the list of primes that
            # divide it. Since we've reached q, we no longer
            # need it in the map, but we'll mark the next
            # multiples of its witnesses to prepare for larger
            # numbers
            #
            for p in d[q]:
                d.setdefault(p + q, []).append(p)
            del d[q]

        q += 1


def prime_encode_structure_list(structure_list):
    """Encode a list of structures using prime labels

    Args:
        structure_list (list [SimpleITK.Image]): A list of binary masks.

    Returns:
        SimpleITK.Image: The prime-encoded structure
    """

    # Create an image with all ones
    img_size = structure_list[0].GetSize()
    prime_encoded_image = sitk.GetImageFromArray(np.ones(img_size[::-1]))
    prime_encoded_image = sitk.Cast(prime_encoded_image, sitk.sitkUInt64)
    prime_encoded_image.CopyInformation(structure_list[0])

    prime_generator = generate_primes()

    for s_img, prime in zip(structure_list, prime_generator):
        # Cast to int
        s_img_int = sitk.Cast(s_img > 0, sitk.sitkUInt64)

        # Multiply with the encoded image
        prime_encoded_image = (
            sitk.Cast(
                sitk.MaskNegated(
                    sitk.Cast(prime_encoded_image, sitk.sitkUInt8),
                    sitk.Cast(s_img_int, sitk.sitkUInt8),
                ),
                sitk.sitkUInt64,
            )
            + sitk.Mask(prime_encoded_image, sitk.Cast(s_img_int, sitk.sitkUInt8))
            * prime
            * s_img_int
        )

    return prime_encoded_image


def prime_decode_image(prime_encoded_image):
    """Decode a prime-encoded image

    Args:
        prime_encoded_image (SimpleITK.Image): A prime-encoded image.

    Returns:
        list [SimpleITK.Image]: A list of binary masks.
    """

    prime_generator = generate_primes()

    structure_list = []
    num_nonzero_voxels = 1

    for prime in prime_generator:
        # Get the next prime
        print(prime)

        # Calculate the region originally defined with this prime
        s_img = sitk.Equal(sitk.Modulus(prime_encoded_image, prime), 0)

        # Check how many voxels we have
        num_nonzero_voxels = sitk.GetArrayViewFromImage(s_img).sum()

        if num_nonzero_voxels > 0:
            structure_list.append(s_img)
        else:
            break

    return structure_list


def binary_encode_structure_list(structure_list):
    """Encode a list of binary labels using binary encoding

    Args:
        structure_list (list (SimpleITK.Image)): The list of binary label maps.

    Raises:
        ValueError: A maximum of 32 structures can be encoded!

    Returns:
        SimpleITK.Image: The encoded image, can be saved etc. as usual.
    """

    if len(structure_list) > 32:
        raise ValueError("You can only encode a maximum of 32 structures with this method!")

    # Create an image with all zeros
    img_size = structure_list[0].GetSize()
    binary_encoded_arr = np.zeros(img_size[::-1]).astype(int)

    for power, s_img in enumerate(structure_list):
        # Convert image to array
        s_arr = sitk.GetArrayFromImage(s_img).astype(int)
        # Bitwise-or with existing array
        binary_encoded_arr = np.bitwise_or(
            binary_encoded_arr, s_arr.astype(bool) * 2 ** (power + 1)
        )

    binary_encoded_img = sitk.GetImageFromArray(binary_encoded_arr)
    binary_encoded_img.CopyInformation(structure_list[0])

    # Image can be cast to 8,16, or 32 bit integer (depending on number of structures)
    # Safest is just to use 32 bits
    binary_encoded_img = sitk.Cast(binary_encoded_img, sitk.sitkUInt32)

    return binary_encoded_img


def binary_decode_image(binary_encoded_img):
    """Decode a binary label map to a list of structures.

    Args:
        binary_encoded_img (SimpleITK.Image): The encoded image.

    Returns:
        list (SimpleITK.Image): The list of images.
    """

    binary_encoded_arr = sitk.GetArrayFromImage(binary_encoded_img).astype(int)

    structure_list = []
    num_nonzero_voxels = 1

    for power in range(32):
        # Calculate the region originally defined with this prime
        s_arr = np.bitwise_and(binary_encoded_arr, 2 ** (power + 1))

        # Check how many voxels we have
        num_nonzero_voxels = s_arr.sum()

        if num_nonzero_voxels > 0:
            s_img = sitk.GetImageFromArray(s_arr) > 0
            s_img.CopyInformation(binary_encoded_img)
            structure_list.append(s_img)
        else:
            continue

    return structure_list



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import pydicom
import SimpleITK as sitk

from platipy.backend import app, DataObject, celery  # pylint: disable=unused-import
from platipy.dicom.io.nifti_to_rtstruct import convert_nifti


logger = logging.getLogger(__name__)

body_settings_defaults = {
    "outputContourName": "primitive_body_contour",
    "seed": [0, 0, 0],
    "lowerThreshold": -5000,
    "upperThreshold": -800,
    "vectorRadius": [1, 1, 1],
}


@app.register("Primitive Body Segmentation", default_settings=body_settings_defaults)
def primitive_body_segmentation(data_objects, working_dir, settings):

    logger.info("Running Primitive Body Segmentation")
    logger.info("Using settings: %s", settings)

    output_objects = []
    for d in data_objects:
        logger.info("Running on data object: %s", d.path)

        # Read the image series
        load_path = d.path
        if d.type == "DICOM":
            load_path = sitk.ImageSeriesReader().GetGDCMSeriesFileNames(d.path)

        img = sitk.ReadImage(load_path)

        # Region growing using Connected Threshold Image Filter
        seg_con = sitk.ConnectedThreshold(
            img,
            seedList=[tuple(settings["seed"])],
            lower=settings["lowerThreshold"],
            upper=settings["upperThreshold"],
        )

        # Clean up the segmentation
        vector_radius = tuple(settings["vectorRadius"])
        kernel = sitk.sitkBall
        seg_clean = sitk.BinaryMorphologicalClosing(seg_con, vector_radius, kernel)
        mask = sitk.BinaryNot(seg_clean)

        # Write the mask to a file in the working_dir
        mask_file = os.path.join(working_dir, "{0}.nii.gz".format(settings["outputContourName"]))
        sitk.WriteImage(mask, mask_file)

        # Create the output Data Object and add it to the list of output_objects
        data_object = DataObject(type="FILE", path=mask_file, parent=d)
        output_objects.append(data_object)

        # If the input was a DICOM, then we can use it to generate an output RTStruct
        if d.type == "DICOM":

            dicom_file = load_path[0]
            logger.info("Will write Dicom using file: %s", dicom_file)
            masks = {settings["outputContourName"]: mask_file}

            # Use the image series UID for the file of the RTStruct
            suid = pydicom.dcmread(dicom_file).SeriesInstanceUID
            output_file = os.path.join(working_dir, "RS.{0}.dcm".format(suid))

            # Use the convert nifti function to generate RTStruct from nifti masks
            convert_nifti(dicom_file, masks, output_file)

            # Create the Data Object for the RTStruct and add it to the list
            do = DataObject(type="DICOM", path=output_file, parent=d)
            output_objects.append(do)

            logger.info("RTStruct generated")

    return output_objects


if __name__ == "__main__":

    # Run app by calling "python sample.py" from the command line

    dicom_listener_port = 7777
    dicom_listener_aetitle = "SAMPLE_SERVICE"

    app.run(
        debug=True,
        host="0.0.0.0",
        port=8000,
        dicom_listener_port=dicom_listener_port,
        dicom_listener_aetitle=dicom_listener_aetitle,
    )



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import SimpleITK as sitk

from platipy.imaging.label.utils import get_com

from platipy.imaging.generation.image import insert_cylinder_image

from platipy.imaging.utils.geometry import vector_angle, rotate_image

from platipy.imaging.utils.crop import crop_to_roi, label_to_roi


def generate_valve_from_great_vessel(
    label_great_vessel,
    label_ventricle,
    valve_thickness_mm=8,
):
    """
    Generates a geometrically-defined valve.
    This function is suitable for the pulmonic and aortic valves.

    Args:
        label_great_vessel (SimpleITK.Image): The binary mask for the great vessel
            (pulmonary artery or ascending aorta)
        label_ventricle (SimpleITK.Image): The binary mask for the ventricle (left or right)
        valve_thickness_mm (int, optional): Valve thickness, in millimetres. Defaults to 8.

    Returns:
        SimpleITK.Image: The geometric valve, as a binary mask.
    """

    # To speed up binary morphology operations we first crop all images
    template_img = 0 * label_ventricle
    cb_size, cb_index = label_to_roi(
        (label_great_vessel + label_ventricle) > 0, expansion_mm=(20, 20, 20)
    )

    label_ventricle = crop_to_roi(label_ventricle, cb_size, cb_index)
    label_great_vessel = crop_to_roi(label_great_vessel, cb_size, cb_index)

    # Convert valve thickness to voxels
    _, _, res_z = label_ventricle.GetSpacing()
    valve_thickness = int(valve_thickness_mm / res_z)

    # Dilate the ventricle
    label_ventricle_dilate = sitk.BinaryDilate(label_ventricle, (valve_thickness,) * 3)

    # Find the overlap
    overlap = label_great_vessel & label_ventricle_dilate

    # Mask to thinner great vessel
    mask = label_great_vessel | label_ventricle_dilate

    overlap = sitk.Mask(overlap, mask)

    label_valve = sitk.BinaryMorphologicalClosing(overlap)

    # Finally, paste back to the original image space
    label_valve = sitk.Paste(
        template_img,
        label_valve,
        label_valve.GetSize(),
        (0, 0, 0),
        cb_index,
    )

    return label_valve


def generate_valve_using_cylinder(
    label_atrium,
    label_ventricle,
    radius_mm=15,
    height_mm=10,
):
    """
    Generates a geometrically-defined valve.
    This function is suitable for the tricuspid and mitral valves.

    Args:
        label_atrium (SimpleITK.Image): The binary mask for the (left or right) atrium.
        label_ventricle (SimpleITK.Image): The binary mask for the (left or right) ventricle.
        radius_mm (int, optional): The valve radius, in mm. Defaults to 15.
        height_mm (int, optional): The valve height (i.e. perpendicular extent), in mm.
            Defaults to 10.

    Returns:
        SimpleITK.Image: The geometrically defined valve
    """
    # To speed up binary morphology operations we first crop all images
    template_img = 0 * label_ventricle
    cb_size, cb_index = label_to_roi(
        (label_atrium + label_ventricle) > 0, expansion_mm=(20, 20, 20)
    )

    label_atrium = crop_to_roi(label_atrium, cb_size, cb_index)
    label_ventricle = crop_to_roi(label_ventricle, cb_size, cb_index)

    # Define the overlap region (using binary dilation)
    # Increment overlap to make sure we have enough voxels
    dilation = 1
    overlap_vol = 0
    while overlap_vol <= 2000:
        dilation_img = [int(dilation / i) for i in label_ventricle.GetSpacing()]
        overlap = sitk.BinaryDilate(label_atrium, dilation_img) & sitk.BinaryDilate(
            label_ventricle, dilation_img
        )
        overlap_vol = np.sum(sitk.GetArrayFromImage(overlap) * np.prod(overlap.GetSpacing()))
        dilation += 1

    # Now we can calculate the location of the valve
    valve_loc = get_com(overlap, as_int=True)
    valve_loc_real = get_com(overlap, real_coords=True)

    # Now we create a cylinder with the user_defined parameters
    cylinder = insert_cylinder_image(0 * label_ventricle, radius_mm, height_mm, valve_loc[::-1])

    # Now we compute the first principal moment (long axis) of the combined chambers
    # f = sitk.LabelShapeStatisticsImageFilter()
    # f.Execute(label_ventricle + label_atrium)
    # orientation_vector = f.GetPrincipalAxes(1)[:3]

    # A more robust method is to use the COM offset from the chambers
    # as a proxy for the long axis of the LV/RV
    orientation_vector = np.array(get_com(label_ventricle, real_coords=True)) - np.array(
        get_com(label_atrium, real_coords=True)
    )

    # Another method is to compute the third principal moment of the overlap region
    # f = sitk.LabelShapeStatisticsImageFilter()
    # f.Execute(overlap)
    # orientation_vector = f.GetPrincipalAxes(1)[:3]

    # Get the rotation parameters
    rotation_angle = vector_angle(orientation_vector, (0, 0, 1), smallest=False)
    rotation_axis = np.cross(orientation_vector, (0, 0, 1))

    # Rotate the cylinder to define the valve
    label_valve = rotate_image(
        cylinder,
        rotation_centre=valve_loc_real,
        rotation_axis=rotation_axis,
        rotation_angle_radians=rotation_angle,
        interpolation=sitk.sitkNearestNeighbor,
        default_value=0,
    )

    # Now we want to trim any parts of the valve too close to the edge of the chambers
    # combined_chambers = sitk.BinaryDilate(label_atrium, (3,) * 3) | sitk.BinaryDilate(
    #     label_ventricle, (3,) * 3
    # )
    # combined_chambers = sitk.BinaryErode(combined_chambers, (6, 6, 6))

    # label_valve = sitk.Mask(label_valve, combined_chambers)

    # Finally, paste back to the original image space
    label_valve = sitk.Paste(
        template_img,
        label_valve,
        label_valve.GetSize(),
        (0, 0, 0),
        cb_index,
    )

    return label_valve



import zipfile
import tempfile
import urllib.request
from pathlib import Path

def download_and_extract_zip_file(zip_url, output_directory):

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = Path(temp_dir).joinpath("tmp.zip")

        with urllib.request.urlopen(zip_url) as dl_file:
            with open(temp_file, "wb") as out_file:
                out_file.write(dl_file.read())

        with zipfile.ZipFile(temp_file, "r") as zip_ref:
            zip_ref.extractall(output_directory)



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import SimpleITK as sitk

from platipy.imaging.registration.utils import (
    smooth_and_resample,
    apply_transform,
    registration_command_iteration,
    stage_iteration,
    deformable_registration_command_iteration,
    control_point_spacing_distance_to_number,
)

from platipy.imaging import ImageVisualiser
from platipy.imaging.label.utils import get_com


def multiscale_demons(
    registration_algorithm,
    fixed_image,
    moving_image,
    initial_transform=None,
    initial_displacement_field=None,
    isotropic_resample=None,
    resolution_staging=None,
    smoothing_sigmas=None,
    iteration_staging=None,
    interp_order=sitk.sitkLinear,
):
    """
    Run the given registration algorithm in a multiscale fashion. The original scale should not be
    given as input as the original images are implicitly incorporated as the base of the pyramid.
    Args:
        registration_algorithm: Any registration algorithm that has an Execute(fixed_image,
                                moving_image, displacement_field_image) method.
        fixed_image: Resulting transformation maps points from this image's spatial domain to the
                     moving image spatial domain.
        moving_image: Resulting transformation maps points from the fixed_image's spatial domain to
                      this image's spatial domain.
        initial_transform: Any SimpleITK transform, used to initialize the displacement field.
        initial_displacement_field: Initial displacement field, if this is provided
                                    initial_transform will be ignored
        shrink_factors: Shrink factors relative to the original image's size.
        smoothing_sigmas: Amount of smoothing which is done prior to resmapling the image using the
                          given shrink factor. These are in physical (image spacing) units.
    Returns:
        SimpleITK.DisplacementFieldTransform
        [Optional] Displacemment (vector) field
    """
    # Create image pyramid.
    fixed_images = []
    moving_images = []

    for resolution, smoothing_sigma in zip(resolution_staging, smoothing_sigmas):
        isotropic_voxel_size_mm = None
        shrink_factor = None

        if isotropic_resample:
            isotropic_voxel_size_mm = resolution
        else:
            shrink_factor = resolution

        fixed_images.append(
            smooth_and_resample(
                fixed_image,
                isotropic_voxel_size_mm=isotropic_voxel_size_mm,
                shrink_factor=shrink_factor,
                smoothing_sigma=smoothing_sigma,
                interpolator=interp_order,
            )
        )

        moving_images.append(
            smooth_and_resample(
                moving_image,
                isotropic_voxel_size_mm=isotropic_voxel_size_mm,
                shrink_factor=shrink_factor,
                smoothing_sigma=smoothing_sigma,
                interpolator=interp_order,
            )
        )

    # Create initial displacement field at lowest resolution.
    # Currently, the pixel type is required to be sitkVectorFloat64 because of a constraint imposed
    # by the Demons filters.
    if not initial_displacement_field:
        if initial_transform:
            initial_displacement_field = sitk.TransformToDisplacementField(
                initial_transform,
                sitk.sitkVectorFloat64,
                fixed_image.GetSize(),
                fixed_image.GetOrigin(),
                fixed_image.GetSpacing(),
                fixed_image.GetDirection(),
            )
        else:
            if len(moving_image.GetSize()) == 2:
                initial_displacement_field = sitk.Image(
                    fixed_image.GetWidth(),
                    fixed_image.GetHeight(),
                    sitk.sitkVectorFloat64,
                )
            elif len(moving_image.GetSize()) == 3:
                initial_displacement_field = sitk.Image(
                    fixed_image.GetWidth(),
                    fixed_image.GetHeight(),
                    fixed_image.GetDepth(),
                    sitk.sitkVectorFloat64,
                )
            initial_displacement_field.CopyInformation(fixed_image)
    else:
        initial_displacement_field = sitk.Resample(initial_displacement_field, fixed_image)

    # Run the registration.
    # Start at the top of the pyramid and work our way down.

    dvf_total = sitk.Resample(initial_displacement_field, fixed_image)

    for i in range(len(fixed_images)):
        f_image = fixed_images[i]
        m_image = moving_images[i]

        # we now apply the (total) transform to the moving image
        dvf_total = sitk.Resample(dvf_total, f_image)

        tfm_total = sitk.DisplacementFieldTransform(sitk.Cast(dvf_total, sitk.sitkVectorFloat64))
        m_image = sitk.Resample(m_image, tfm_total, interp_order)

        # set up iteration staging
        iters = iteration_staging[i]
        registration_algorithm.SetNumberOfIterations(iters)

        # set up regularisation
        # leave constant (in image units) for now

        dvf_iter = registration_algorithm.Execute(f_image, m_image)

        # and now add to the running DVF
        # importly, at each voxel the deformation vector (source -> destination)
        # has to be updated with the vector field itself
        dvf_total = dvf_total + sitk.Resample(dvf_iter, tfm_total)

        # manually smooth the DVF
        sigma = registration_algorithm.GetStandardDeviations()
        dvf_total = sitk.SmoothingRecursiveGaussian(dvf_total, sigma)
        dvf_total = sitk.Cast(dvf_total, sitk.sitkVectorFloat64)

        # vis = ImageVisualiser(f_image, cut=get_com(f_image), figure_size_in=6)
        # vis.add_comparison_overlay(m_image)

        # vis.add_vector_overlay(
        #     dvf_iter,
        #     arrow_scale=0.25,
        #     arrow_width=0.25,
        #     subsample=4,
        # )

        # vis.set_limits_from_label(f_image, expansion=100)
        # fig = vis.show()

        # test_tfm = sitk.DisplacementFieldTransform(
        #     sitk.Cast(dvf_iter, sitk.sitkVectorFloat64)
        # )
        # test = sitk.Resample(m_image, test_tfm)

        # vis = ImageVisualiser(f_image > 0, cut=get_com(f_image), figure_size_in=6)
        # vis.add_comparison_overlay(test > 0)

        # vis.set_limits_from_label(f_image, expansion=100)
        # fig = vis.show()

    dvf_total = sitk.Resample(dvf_total, fixed_image)

    return dvf_total


def fast_symmetric_forces_demons_registration(
    fixed_image,
    moving_image,
    resolution_staging=[8, 4, 1],
    iteration_staging=[10, 10, 10],
    isotropic_resample=False,
    initial_displacement_field=None,
    regularisation_kernel_mm=1.5,
    smoothing_sigma_factor=1,
    smoothing_sigmas=False,
    default_value=None,
    ncores=1,
    interp_order=2,
    verbose=False,
):
    """
    Deformable image propagation using Fast Symmetric-Forces Demons

    Args
        fixed_image (sitk.Image)        : the fixed image
        moving_image (sitk.Image)       : the moving image, to be deformable registered (must be in
                                          the same image space)
        resolution_staging (list[int])   : down-sampling factor for each resolution level
        iteration_staging (list[int])    : number of iterations for each resolution level
        isotropic_resample (bool)        : flag to request isotropic resampling of images, in which
                                           case resolution_staging is used to define voxel size
                                           (mm) per level
        initial_displacement_field (sitk.Image) : Initial displacement field to use
        regularisation_kernel_scale (float) : Relative scale (var/voxel size) of the regularisation kernel (Gaussian)
        ncores (int)                    : number of processing cores to use
        smoothing_sigma_factor (float)    : the relative width of the Gaussian smoothing kernel
        interp_order (int)               : the interpolation order
                                            1 = Nearest neighbour
                                            2 = Bi-linear splines
                                            3 = B-Spline (cubic)

        default_value (float) : Default voxel value. Defaults to 0 unless image is CT-like.

    Returns
        registered_image (sitk.Image)    : the registered moving image
        output_transform                 : the displacement field transform
        [optional] deformation_field
    """

    # Cast to floating point representation, if necessary

    moving_image_type = moving_image.GetPixelID()

    if fixed_image.GetPixelID() != 6:
        fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    if moving_image.GetPixelID() != 6:
        moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # Set up the appropriate image filter
    registration_method = sitk.FastSymmetricForcesDemonsRegistrationFilter()

    # Multi-resolution framework
    registration_method.SetNumberOfThreads(ncores)
    registration_method.SetSmoothUpdateField(True)
    registration_method.SetSmoothDisplacementField(True)

    # This is the regularisation kernel
    # values are set in image (voxel) coordinates
    regularisation_kernel_vox = np.array(regularisation_kernel_mm) / np.array(
        fixed_image.GetSpacing()
    )

    registration_method.SetStandardDeviations(regularisation_kernel_vox.tolist())

    # This allows monitoring of the progress
    if verbose:
        registration_method.AddCommand(
            sitk.sitkIterationEvent,
            lambda: deformable_registration_command_iteration(registration_method),
        )

    if not smoothing_sigmas:
        smoothing_sigmas = [i * smoothing_sigma_factor for i in resolution_staging]

    deformation_field = multiscale_demons(
        registration_algorithm=registration_method,
        fixed_image=fixed_image,
        moving_image=moving_image,
        resolution_staging=resolution_staging,
        smoothing_sigmas=smoothing_sigmas,
        iteration_staging=iteration_staging,
        isotropic_resample=isotropic_resample,
        initial_displacement_field=initial_displacement_field,
        interp_order=interp_order,
    )

    resampler = sitk.ResampleImageFilter()
    resampler.SetReferenceImage(fixed_image)
    resampler.SetInterpolator(interp_order)

    # Try to find default value
    if default_value is None:
        default_value = 0

        # Test if image is CT-like
        if sitk.GetArrayViewFromImage(moving_image).min() <= -1000:
            default_value = -1000

    resampler.SetDefaultPixelValue(default_value)

    # create the deformable transform
    output_transform = sitk.DisplacementFieldTransform(
        sitk.Cast(deformation_field, sitk.sitkVectorFloat64)
    )

    resampler.SetTransform(output_transform)
    registered_image = resampler.Execute(moving_image)

    registered_image.CopyInformation(fixed_image)
    registered_image = sitk.Cast(registered_image, moving_image_type)

    return registered_image, output_transform, deformation_field


def bspline_registration(
    fixed_image,
    moving_image,
    fixed_structure=False,
    moving_structure=False,
    resolution_staging=[8, 4, 2],
    smooth_sigmas=[4, 2, 1],
    sampling_rate=0.1,
    optimiser="LBFGS",
    metric="mean_squares",
    initial_grid_spacing=64,
    grid_scale_factors=[1, 2, 4],
    interp_order=3,
    default_value=-1000,
    number_of_iterations=20,
    isotropic_resample=False,
    initial_isotropic_size=1,
    number_of_histogram_bins_mi=30,
    verbose=False,
    ncores=8,
):
    """
    B-Spline image registration using ITK

    IMPORTANT - THIS IS UNDER ACTIVE DEVELOPMENT

    Args:
        fixed_image ([SimpleITK.Image]): The fixed (target/primary) image.
        moving_image ([SimpleITK.Image]): The moving (secondary) image.
        fixed_structure (bool, optional): If defined, a binary SimpleITK.Image used to mask metric
                                          evaluation for the moving image. Defaults to False.
        moving_structure (bool, optional): If defined, a binary SimpleITK.Image used to mask metric
                                           evaluation for the fixed image. Defaults to False.
        resolution_staging (list, optional): The multi-resolution downsampling factors.
                                             Defaults to [8, 4, 2].
        smooth_sigmas (list, optional): The multi-resolution smoothing kernel scale (Gaussian).
                                        Defaults to [4, 2, 1].
        sampling_rate (float, optional): The fraction of voxels sampled during each iteration.
                                         Defaults to 0.1.
        optimiser (str, optional): The optimiser algorithm used for image registration.
                                   Available options:
                                    - LBFSGS
                                      (limited-memory Broyden–Fletcher–Goldfarb–Shanno (bounded).)
                                    - LBFSG
                                      (limited-memory Broyden–Fletcher–Goldfarb–Shanno
                                      (unbounded).)
                                    - CGLS (conjugate gradient line search)
                                    - gradient_descent
                                    - gradient_descent_line_search
                                   Defaults to "LBFGS".
        metric (str, optional): The metric to be optimised during image registration.
                                Available options:
                                 - correlation
                                 - mean_squares
                                 - demons
                                 - mutual_information
                                   (used with parameter number_of_histogram_bins_mi)
                                Defaults to "mean_squares".
        initial_grid_spacing (int, optional): Grid spacing of lower resolution stage (in mm).
                                              Defaults to 64.
        grid_scale_factors (list, optional): Factors to determine grid spacing at each
                                             multiresolution stage.
                                             Defaults to [1, 2, 4].
        interp_order (int, optional): Interpolation order of final resampling.
                                      Defaults to 3 (cubic).
        default_value (int, optional): Default image value. Defaults to -1000.
        number_of_iterations (int, optional): Number of iterations at each resolution stage.
                                              Defaults to 20.
        isotropic_resample (bool, optional): Flag whether to resample to isotropic resampling
                                             prior to registration.
                                             Defaults to False.
        initial_isotropic_size (int, optional): Voxel size (in mm) of resampled isotropic image
                                                (if used). Defaults to 1.
        number_of_histogram_bins_mi (int, optional): Number of histogram bins used when calculating
                                                     mutual information. Defaults to 30.
        verbose (bool, optional): Print image registration process information. Defaults to False.
        ncores (int, optional): Number of CPU cores used. Defaults to 8.

    Returns:
        [SimpleITK.Image]: The registered moving (secondary) image.
        [SimleITK.Transform]: The linear transformation.

    Notes:
     - smooth_sigmas are relative to resolution staging
        e.g. for image spacing of 1x1x1 mm^3, with smooth sigma=2 and resolution_staging=4, the
        scale of the Gaussian filter would be 2x4 = 8mm (i.e. 8x8x8 mm^3)

    """

    # Re-cast input images
    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)

    moving_image_type = moving_image.GetPixelID()
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # (Optional) isotropic resample
    # This changes the behaviour, so care should be taken
    # For highly anisotropic images may be preferable

    if isotropic_resample:
        # First, copy the fixed image so we can resample back into this space at the end
        fixed_image_original = fixed_image
        fixed_image_original.MakeUnique()

        fixed_image = smooth_and_resample(
            fixed_image,
            isotropic_voxel_size_mm=initial_isotropic_size,
        )

        moving_image = smooth_and_resample(
            moving_image,
            isotropic_voxel_size_mm=initial_isotropic_size,
        )

    else:
        fixed_image_original = fixed_image

    # Set up image registration method
    registration = sitk.ImageRegistrationMethod()
    registration.SetNumberOfThreads(ncores)

    registration.SetShrinkFactorsPerLevel(resolution_staging)
    registration.SetSmoothingSigmasPerLevel(smooth_sigmas)
    registration.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

    # Choose optimiser
    if optimiser.lower() == "lbfgsb":
        registration.SetOptimizerAsLBFGSB(
            gradientConvergenceTolerance=1e-5,
            numberOfIterations=number_of_iterations,
            maximumNumberOfCorrections=5,
            maximumNumberOfFunctionEvaluations=1024,
            costFunctionConvergenceFactor=1e7,
            trace=verbose,
        )
    elif optimiser.lower() == "lbfgs":
        registration.SetOptimizerAsLBFGS2(
            numberOfIterations=number_of_iterations,
            solutionAccuracy=1e-2,
            hessianApproximateAccuracy=6,
            deltaConvergenceDistance=0,
            deltaConvergenceTolerance=0.01,
            lineSearchMaximumEvaluations=40,
            lineSearchMinimumStep=1e-20,
            lineSearchMaximumStep=1e20,
            lineSearchAccuracy=0.01,
        )
    elif optimiser.lower() == "cgls":
        registration.SetOptimizerAsConjugateGradientLineSearch(
            learningRate=0.05, numberOfIterations=number_of_iterations
        )
        registration.SetOptimizerScalesFromPhysicalShift()
    elif optimiser.lower() == "gradient_descent":
        registration.SetOptimizerAsGradientDescent(
            learningRate=5.0,
            numberOfIterations=number_of_iterations,
            convergenceMinimumValue=1e-6,
            convergenceWindowSize=10,
        )
        registration.SetOptimizerScalesFromPhysicalShift()
    elif optimiser.lower() == "gradient_descent_line_search":
        registration.SetOptimizerAsGradientDescentLineSearch(
            learningRate=1.0, numberOfIterations=number_of_iterations
        )
        registration.SetOptimizerScalesFromPhysicalShift()

    # Set metric
    if metric == "correlation":
        registration.SetMetricAsCorrelation()
    elif metric == "mean_squares":
        registration.SetMetricAsMeanSquares()
    elif metric == "demons":
        registration.SetMetricAsDemons()
    elif metric == "mutual_information":
        registration.SetMetricAsMattesMutualInformation(
            numberOfHistogramBins=number_of_histogram_bins_mi
        )

    registration.SetInterpolator(sitk.sitkLinear)

    # Set sampling
    if isinstance(sampling_rate, float):
        registration.SetMetricSamplingPercentage(sampling_rate)
    elif type(sampling_rate) in [np.ndarray, list]:
        registration.SetMetricSamplingPercentagePerLevel(sampling_rate)

    registration.SetMetricSamplingStrategy(sitk.ImageRegistrationMethod.REGULAR)

    # Set masks
    if moving_structure is not False:
        registration.SetMetricMovingMask(moving_structure)

    if fixed_structure is not False:
        registration.SetMetricFixedMask(fixed_structure)

    # Set control point spacing
    transform_domain_mesh_size = control_point_spacing_distance_to_number(
        fixed_image, initial_grid_spacing
    )

    if verbose:
        print(f"Initial grid size: {transform_domain_mesh_size}")

    # Initialise transform
    initial_transform = sitk.BSplineTransformInitializer(
        fixed_image,
        transformDomainMeshSize=[int(i) for i in transform_domain_mesh_size],
    )
    registration.SetInitialTransformAsBSpline(
        initial_transform, inPlace=True, scaleFactors=grid_scale_factors
    )

    # (Optionally) add iteration commands
    if verbose:
        registration.AddCommand(
            sitk.sitkIterationEvent,
            lambda: registration_command_iteration(registration),
        )
        registration.AddCommand(
            sitk.sitkMultiResolutionIterationEvent,
            lambda: stage_iteration(registration),
        )

    # Run the registration
    output_transform = registration.Execute(fixed=fixed_image, moving=moving_image)

    # Resample moving image
    registered_image = apply_transform(
        input_image=moving_image,
        reference_image=fixed_image_original,
        transform=output_transform,
        default_value=default_value,
        interpolator=interp_order,
    )

    registered_image = sitk.Cast(registered_image, moving_image_type)

    # Return outputs
    return registered_image, output_transform



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

import logging

from platipy.utils import download_and_extract_zip_file

logger = logging.getLogger(__name__)

LCTSC_TEST_DATA_URL = "https://zenodo.org/record/4747795/files"
LCTSC_DICOM_ZIP = "LCTSC_DICOM_TestData.zip"
LCTSC_NIFTI_ZIP = "LCTSC_NIFTI_TestData.zip"

TCGA_HNSC_TEST_DATA_URL = "https://zenodo.org/record/5147890/files"
TCGA_HNSC_NIFTI_ZIP = "/TCGA-HNSC_NIFTI_TestData.zip"


def get_lung_dicom(output_directory="./data/dicom"):
    """Get some Lung data for testing in DICOM format. If the output_directory already exists no
    data will be fetched.

    Args:
        output_directory (str, optional): Path to store test Lung data. Defaults to "./data/dicom".

    Returns:
        str: The output directory containing the downloaded data.
    """

    output_directory = Path(output_directory)

    if output_directory.exists():
        logger.debug("Output directory exists, stopping. %s", output_directory)
        return output_directory

    zip_url = f"{LCTSC_TEST_DATA_URL}/{LCTSC_DICOM_ZIP}?download=1"
    download_and_extract_zip_file(zip_url, output_directory)

    return output_directory


def get_lung_nifti(output_directory="./data/nifti/lung"):
    """Get some Lung data for testing in Nifti format. If the output_directory already exists no
    data will be fetched.

    Args:
        output_directory (str, optional): Path to store test Lung data. Defaults to
          "./data/nifti/lung".

    Returns:
        str: The output directory containing the downloaded data.
    """

    output_directory = Path(output_directory)

    if output_directory.exists():
        logger.debug("Output directory exists, stopping. %s", output_directory)
        return output_directory

    zip_url = f"{LCTSC_TEST_DATA_URL}/{LCTSC_NIFTI_ZIP}?download=1"
    download_and_extract_zip_file(zip_url, output_directory)

    return output_directory


def get_hn_nifti(output_directory="./data/nifti/hn"):
    """Get some Head and Neck data for testing in Nifti format. If the output_directory already
    exists no data will be fetched.

    Args:
        output_directory (str, optional): Path to store test Lung data. Defaults to
          "./data/nifti/hn".

    Returns:
        str: The output directory containing the downloaded data.
    """

    output_directory = Path(output_directory)

    if output_directory.exists():
        logger.debug("Output directory exists, stopping. %s", output_directory)
        return output_directory

    zip_url = f"{TCGA_HNSC_TEST_DATA_URL}/{TCGA_HNSC_NIFTI_ZIP}?download=1"
    download_and_extract_zip_file(zip_url, output_directory)

    return output_directory



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
import os
import datetime
import shutil
import time
import logging

from celery.schedules import crontab

from platipy.backend import celery, db, app
from platipy.dicom.communication import DicomConnector

from .models import Dataset, DataObject

logger = logging.getLogger(__name__)

celery.conf.beat_schedule = {
    # Executes every hour
    "clean-up-every-hour": {
        "task": "platipy.backend.tasks.clean_up_task",
        "schedule": crontab(minute=0),
        "args": (),
    },
}

celery.conf.timezone = "UTC"


@celery.task(bind=True)
def run_dicom_listener(task):
    app.run_dicom_listener()


@celery.task(bind=True)
def clean_up_task(task):
    """
    Deletes DataObjects from expired Datasets
    """

    logger.info("Running Clean Up Task")

    datasets = Dataset.query.all()

    now = datetime.datetime.now()

    num_data_objs_removed = 0

    for ds in datasets:

        if ds.timeout < now:

            data_objects = ds.input_data_objects + ds.output_data_objects

            for do in data_objects:

                try:
                    if do.path:
                        if os.path.isdir(do.path):
                            logger.debug("Removing Directory: %s", do.path)
                            shutil.rmtree(do.path)
                        elif os.path.isfile(do.path):
                            logger.debug("Removing File: %s", do.path)
                            os.remove(do.path)
                        else:
                            logger.debug(
                                "Data missing, must have already been deleted: %s",
                                do.path,
                            )

                        num_data_objs_removed += 1

                        do.is_fetched = False
                        do.path = None
                        db.session.commit()

                except Exception as e:
                    logger.warning("Exception occured when removing DataObject: %s", do)

    logger.info("Clean Up Task Complete: Removed %s DataObjects", num_data_objs_removed)


@celery.task(bind=True)
def retrieve_task(task, data_object_id):
    """
    Fetch a Dicom Object from the dicom location using the retrieve_type GET or MOVE
    """

    do = DataObject.query.filter_by(id=data_object_id).first()

    dicom_connector = DicomConnector(
        host=do.dataset.from_dicom_location.host,
        port=do.dataset.from_dicom_location.port,
        ae_title=do.dataset.from_dicom_location.ae_title,
    )
    dicom_verify = dicom_connector.verify()

    if not dicom_verify:
        logger.error(
            "Unable to connect to Dicom Location: %s %s %s",
            do.dataset.from_dicom_location.host,
            do.dataset.from_dicom_location.port,
            do.dataset.from_dicom_location.ae_title,
        )
        return

    dicom_path = dicom_connector.download_series(do.series_instance_uid)

    do.is_fetched = True
    do.path = dicom_path
    db.session.commit()


@celery.task(bind=True)
def move_task(task, endpoint, seriesUIDs, host, port, ae_title):
    """
    Background task that triggers the Dicom MOVE operation at the given endpoint
    for the given seriesUIDs
    """

    # For each series UID supplied, fetch the image series and run the algorithm
    total = len(seriesUIDs)
    count = 0

    dicom_connector = DicomConnector(host=host, port=port, ae_title=ae_title)

    task.update_state(
        state="PROGRESS",
        meta={"current": count, "total": total, "status": "Verifying dicom location"},
    )

    dicom_verify = dicom_connector.verify()

    if dicom_verify is None:
        return {
            "current": 100,
            "total": 100,
            "status": "Unable to connect to dicom location",
        }

    for suid in seriesUIDs:

        task.update_state(
            state="PROGRESS",
            meta={
                "current": count,
                "total": total,
                "status": f"Moving series for UID: {suid}",
            },
        )

        logger.info("Moving Series with UID: %s", suid)
        dicom_connector.move_series(suid)

        count = count + 1

    task.update_state(
        state="SUCCESS",
        meta={"current": total, "total": total, "status": "Move Complete"},
    )


@celery.task(bind=True)
def run_task(task, algorithm_name, config, dataset_id):

    task_id = task.request.id

    start = time.time()

    # Commit to refresh session
    db.session.commit()

    algorithm = app.algorithms[algorithm_name]

    if not config:
        config = algorithm.default_settings

    ds = Dataset.query.filter_by(id=dataset_id).first()
    input_objects = ds.input_data_objects

    logger.info("Will run algorithm: %s", algorithm_name)
    logger.info("Using settings: %s", config)
    logger.info("Number of data objects in dataset: %d", len(input_objects))

    state_details = {
        "current": 0,
        "total": len(input_objects),
        "status": f"Running Algorithm: {algorithm_name}",
    }

    task.update_state(state="RUNNING", meta=state_details)

    if config is None:
        output_data_objects = algorithm.function(input_objects, tempfile.mkdtemp())
    else:
        output_data_objects = algorithm.function(input_objects, tempfile.mkdtemp(), config)

    if not output_data_objects:
        logger.warning("Algorithm (%s) did not return any output objects", algorithm_name)

    # Save the data objects
    for do in output_data_objects:
        do.dataset_id = ds.id
        db.session.add(do)
        db.session.commit()

        if do.type == "DICOM":
            if ds.to_dicom_location:

                logger.info("Sending to Dicom To Location")
                dicom_connector = DicomConnector(
                    host=do.dataset.to_dicom_location.host,
                    port=do.dataset.to_dicom_location.port,
                    ae_title=do.dataset.to_dicom_location.ae_title,
                )
                dicom_verify = dicom_connector.verify()

                if not dicom_verify:
                    logger.error(
                        "Unable to connect to Dicom Location: %s %s %s",
                        do.dataset.to_dicom_location.host,
                        do.dataset.to_dicom_location.port,
                        do.dataset.to_dicom_location.ae_title,
                    )
                    continue

                send_result = dicom_connector.send_dcm(do.path)

                if send_result:
                    do.is_sent = True
                    db.session.add(do)
                    db.session.commit()

            else:
                logger.warning(
                    "DICOM Data Object output but not Dicom To location defined in Dataset"
                )

    end = time.time()
    time_taken = end - start
    logger.info("Dataset processing complete, took: %.2f", time_taken)
    logger.info("Number of data objects generated: %d", len(output_data_objects))

    state_details = {
        "current": len(input_objects),
        "total": len(input_objects),
        "status": f"Running Algorithm Complete: {algorithm_name}",
    }

    task.update_state(state="COMPLETE", meta=state_details)



from matplotlib import cm
import numpy as np
import SimpleITK as sitk

import matplotlib
import matplotlib.colors as mcolors


def write_nrrd_structure_set(
    masks,
    output_file="structure_set.nrrd",
    colormap=matplotlib.colormaps.get_cmap("rainbow"),
    use_compression=True,
    compression_level=9,
):
    """Write an nrrd structure given a set of masks. Useful for reading in Slicer.

    Args:
        masks (dict): Key is structure name, value is SimpleITK.Image
        output_file (str, optional): The path to write the nrrd file. Defaults to
            "structure_set.nrrd".
        colormap (matplotlib.colors.Colormap | dict, optional): Colormap to use for output.
            Can also use a dictionary with keys as structure names and values as colors.
            Defaults to matplotlib.colormaps.get_cmap("rainbow").
        use_compression (bool, optional): Used by SimpleITK's WriteImage function to enable
            compression when writing output file. Defaults to True.
        compression_level (int, optional): Used by SimpleITK's WriteImage function to specify how
            much compression to use. Should be an integer between 0-9. Defaults to 9.

    Raises:
        AttributeError: masks must be a dict
        ValueError: colormap must be a matplotlib colormap or a dictionary
    """

    if not isinstance(masks, dict):
        raise AttributeError("masks should be dict with key name and value SimpleITK.Image")

    all_arr = np.zeros(0)
    structure_count = 1
    structure_names = []
    structure_values = []
    structure_layers = []

    for structure_name in masks:
        mask = masks[structure_name]

        arr = sitk.GetArrayFromImage(mask)
        if not len(np.unique(arr)) == 2:
            continue

        arr = arr[:, :, :, np.newaxis]

        structure_value = 1
        dim = 0
        if len(all_arr) > 0:
            # Check if this structure has space in the current dim
            # Append a new dim if not
            struct_added = False
            for dim in range(all_arr.shape[3]):
                if np.count_nonzero(all_arr[:, :, :, dim][arr[:, :, :, 0] > 0]) > 0:
                    # Overlap! look at next dim
                    continue

                structure_value = np.max(all_arr[:, :, :, dim]) + 1
                all_arr[:, :, :, dim] = all_arr[:, :, :, dim] + (
                    arr[:, :, :, 0] * (structure_value)
                )
                struct_added = True
                break

            if not struct_added:
                all_arr = np.concatenate((all_arr, arr), axis=3)
                dim += 1
        else:
            all_arr = arr

        structure_names.append(structure_name)
        structure_values.append(structure_value)
        structure_layers.append(dim)
        structure_count += 1

    struct_im = sitk.GetImageFromArray(all_arr)
    struct_im.SetSpacing(mask.GetSpacing())
    struct_im.SetOrigin(mask.GetOrigin())
    struct_im.SetDirection(mask.GetDirection())
    extent = (
        f"0 {struct_im.GetSize()[0]-1} 0 {struct_im.GetSize()[1]-1} 0 {struct_im.GetSize()[2]-1}"
    )

    structure_count = 0

    for name in structure_names:
        if isinstance(colormap, (mcolors.ListedColormap, mcolors.LinearSegmentedColormap)):
            color = colormap(hash(name) % 256)
        elif isinstance(colormap, dict):
            color = colormap[name]
        else:
            raise ValueError("colormap must be either a matplotlib colormap or a dictionary!")

        color = color[:3]
        color = [str(c) for c in color]

        struct_im.SetMetaData("Segment{0}_Color".format(structure_count), " ".join(color))
        struct_im.SetMetaData("Segment{0}_ColorAutoGenerated".format(structure_count), "1")
        struct_im.SetMetaData("Segment{0}_Extent".format(structure_count), extent)
        struct_im.SetMetaData(
            "Segment{0}_ID".format(structure_count), f"Segment_{structure_count+1}"
        )
        struct_im.SetMetaData(
            "Segment{0}_LabelValue".format(structure_count),
            str(structure_values[structure_count]),
        )
        struct_im.SetMetaData(
            "Segment{0}_Layer".format(structure_count), str(structure_layers[structure_count])
        )
        struct_im.SetMetaData("Segment{0}_Name".format(structure_count), str(name))
        struct_im.SetMetaData("Segment{0}_NameAutoGenerated".format(structure_count), "1")
        struct_im.SetMetaData("Segment{0}_Tags".format(structure_count), "")

        structure_count += 1

    sitk.WriteImage(
        struct_im,
        str(output_file),
        useCompression=use_compression,
        compressionLevel=compression_level,
    )



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

import SimpleITK as sitk
import logging

logger = logging.getLogger(__name__)

# import pydicom

# Need include celery here to be able to from Docker container
# pylint: disable=unused-import
from platipy.backend import app, DataObject, celery

from platipy.imaging.projects.cardiac.run import (
    CARDIAC_SETTINGS_DEFAULTS,
    run_cardiac_segmentation,
)


@app.register("Cardiac Segmentation", default_settings=CARDIAC_SETTINGS_DEFAULTS)
def cardiac_service(data_objects, working_dir, settings):
    """
    Implements the platipy framework to provide cardiac atlas based segmentation.
    """

    logger.info("Running Cardiac Segmentation")
    logger.info("Using settings: %s", settings)

    output_objects = []
    for data_object in data_objects:
        logger.info("Running on data object: %s", data_object.path)

        # Read the image series
        load_path = data_object.path
        if data_object.type == "DICOM":
            load_path = sitk.ImageSeriesReader().GetGDCMSeriesFileNames(data_object.path)

        img = sitk.ReadImage(load_path)

        results, _ = run_cardiac_segmentation(img, settings=settings)

        # Save resulting masks and add to output for service
        for output in results:

            mask_file = os.path.join(working_dir, "{0}.nii.gz".format(output))
            sitk.WriteImage(results[output], mask_file)

            output_data_object = DataObject(type="FILE", path=mask_file, parent=data_object)
            output_objects.append(output_data_object)

        # If the input was a DICOM, then we can use it to generate an output RTStruct
        # if data_object.type == "DICOM":

        #     dicom_file = load_path[0]
        #     logger.info("Will write Dicom using file: {0}".format(dicom_file))
        #     masks = {settings["outputContourName"]: mask_file}

        #     # Use the image series UID for the file of the RTStruct
        #     suid = pydicom.dcmread(dicom_file).SeriesInstanceUID
        #     output_file = os.path.join(working_dir, "RS.{0}.dcm".format(suid))

        #     # Use the convert nifti function to generate RTStruct from nifti masks
        #     convert_nifti(dicom_file, masks, output_file)

        #     # Create the Data Object for the RTStruct and add it to the list
        #     do = DataObject(type="DICOM", path=output_file, parent=d)
        #     output_objects.append(do)

        #     logger.info("RTStruct generated")

    return output_objects


@app.register(
    "Cardiac Structure Guided Segmentation",
    default_settings=CARDIAC_SETTINGS_DEFAULTS,
)
def cardiac_structure_guided_service(data_objects, working_dir, settings):
    """Runs the structure guided cardiac segmentation service"""

    logger.info("Running Structure Guided Cardiac Segmentation")
    logger.info("Using settings: %s", settings)

    output_objects = []
    for data_object in data_objects:
        logger.info("Running on data object: %s", data_object.path)

        # Read the image series
        load_path = data_object.path
        if data_object.type == "DICOM":
            load_path = sitk.ImageSeriesReader().GetGDCMSeriesFileNames(data_object.path)

        img = sitk.ReadImage(load_path)

        # Load the WHOLEHEART contour (child of Image)
        if len(data_object.children) == 0:
            logger.error(
                "Wholeheart structure needed for structure guided cardiac "
                "segmentation, skipping %s",
                data_object.id,
            )
            continue

        wholeheart = sitk.ReadImage(data_object.children[0].path)

        results, _ = run_cardiac_segmentation(img, wholeheart, settings)

        # Save resulting masks and add to output for service
        for output in results:

            mask_file = os.path.join(working_dir, "{0}.nii.gz".format(output))
            sitk.WriteImage(results[output], mask_file)

            output_data_object = DataObject(type="FILE", path=mask_file, parent=data_object)
            output_objects.append(output_data_object)

    return output_objects


if __name__ == "__main__":

    # Run app by calling "python sample.py" from the command line

    DICOM_LISTENER_PORT = 7777
    DICOM_LISTENER_AETITLE = "SAMPLE_SERVICE"

    app.run(
        debug=True,
        host="0.0.0.0",
        port=8000,
        dicom_listener_port=DICOM_LISTENER_PORT,
        dicom_listener_aetitle=DICOM_LISTENER_AETITLE,
    )



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import SimpleITK as sitk

from platipy.imaging.registration.utils import (
    apply_transform,
    registration_command_iteration,
)


def alignment_registration(fixed_image, moving_image, moments=True):
    """
    A simple registration procedure that can align images in a single step.
    Uses the image centres-of-mass (and optionally second moments) to
    estimate the shift (and rotation) needed for alignment.

    Args:
        fixed_image ([SimpleITK.Image]): The fixed (target/primary) image.
        moving_image ([SimpleITK.Image]): The moving (secondary) image.
        moments (bool, optional): Option to align images using the second moment. Defaults to True.

    Returns:
        [SimpleITK.Image]: The registered moving (secondary) image.
        [SimleITK.Transform]: The linear transformation.
    """

    moving_image_type = moving_image.GetPixelIDValue()
    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)
    initial_transform = sitk.CenteredTransformInitializer(
        fixed_image, moving_image, sitk.VersorRigid3DTransform(), moments
    )
    aligned_image = sitk.Resample(moving_image, fixed_image, initial_transform)
    aligned_image = sitk.Cast(aligned_image, moving_image_type)
    return aligned_image, initial_transform


def linear_registration(
    fixed_image,
    moving_image,
    fixed_structure=None,
    moving_structure=None,
    reg_method="similarity",
    metric="mean_squares",
    optimiser="gradient_descent",
    shrink_factors=[8, 2, 1],
    smooth_sigmas=[4, 2, 0],
    sampling_rate=0.25,
    final_interp=2,
    number_of_iterations=50,
    default_value=None,
    verbose=False,
):
    """
    Initial linear registration between two images.
    The images are not required to be in the same space.
    There are several transforms available, with options for the metric and optimiser to be used.
    Note the default_value, which should be set to match the image modality.

    Args:
        fixed_image ([SimpleITK.Image]): The fixed (target/primary) image.
        moving_image ([SimpleITK.Image]): The moving (secondary) image.
        fixed_structure (bool, optional): If defined, a binary SimpleITK.Image used to mask metric
                                          evaluation for the moving image. Defaults to False.
        moving_structure (bool, optional): If defined, a binary SimpleITK.Image used to mask metric
                                           evaluation for the fixed image. Defaults to False.
        reg_method (str, optional): The linear transformtation model to be used for image
                                    registration.
                                    Available options:
                                     - translation
                                     - rigid
                                     - similarity
                                     - affine
                                     - scale
                                     - scaleversor
                                     - scaleskewversor
                                    Defaults to "Similarity".
        metric (str, optional): The metric to be optimised during image registration.
                                Available options:
                                 - correlation
                                 - mean_squares
                                 - mattes_mi
                                 - joint_hist_mi
                                Defaults to "mean_squares".
        optimiser (str, optional): The optimiser algorithm used for image registration.
                                   Available options:
                                    - lbfgsb
                                      (limited-memory Broyden–Fletcher–Goldfarb–Shanno (bounded).)
                                    - gradient_descent
                                    - gradient_descent_line_search
                                   Defaults to "gradient_descent".
        shrink_factors (list, optional): The multi-resolution downsampling factors.
                                         Defaults to [8, 2, 1].
        smooth_sigmas (list, optional): The multi-resolution smoothing kernel scale (Gaussian).
                                        Defaults to [4, 2, 0].
        sampling_rate (float, optional): The fraction of voxels sampled during each iteration.
                                         Defaults to 0.25.
        ants_radius (int, optional): Used is the metric is set as "ants_radius". Defaults to 3.
        final_interp (int, optional): The final interpolation order. Defaults to 2 (linear).
        number_of_iterations (int, optional): Number of iterations in each multi-resolution step.
                                              Defaults to 50.
        default_value (int, optional): Default voxel value. Defaults to 0 unless image is CT-like.
        verbose (bool, optional): Print image registration process information. Defaults to False.

    Returns:
        [SimpleITK.Image]: The registered moving (secondary) image.
        [SimleITK.Transform]: The linear transformation.
    """

    # Re-cast
    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)

    moving_image_type = moving_image.GetPixelIDValue()
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # Initialise using a VersorRigid3DTransform
    initial_transform = sitk.CenteredTransformInitializer(
        fixed_image, moving_image, sitk.Euler3DTransform(), False
    )
    # Set up image registration method
    registration = sitk.ImageRegistrationMethod()

    registration.SetShrinkFactorsPerLevel(shrink_factors)
    registration.SetSmoothingSigmasPerLevel(smooth_sigmas)
    registration.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

    registration.SetMovingInitialTransform(initial_transform)

    if metric.lower() == "correlation":
        registration.SetMetricAsCorrelation()
    elif metric.lower() == "mean_squares":
        registration.SetMetricAsMeanSquares()
    elif metric.lower() == "mattes_mi":
        registration.SetMetricAsMattesMutualInformation()
    elif metric.lower() == "joint_hist_mi":
        registration.SetMetricAsJointHistogramMutualInformation()
    # to do: add the rest

    registration.SetInterpolator(sitk.sitkLinear)  # Perhaps a small gain in improvement
    registration.SetMetricSamplingPercentage(sampling_rate, seed=42)
    registration.SetMetricSamplingStrategy(sitk.ImageRegistrationMethod.REGULAR)

    # This is only necessary if using a transform comprising changes with different units
    # e.g. rigid (rotation: radians, translation: mm)
    # It can safely be left on
    registration.SetOptimizerScalesFromPhysicalShift()

    if moving_structure:
        registration.SetMetricMovingMask(moving_structure)

    if fixed_structure:
        registration.SetMetricFixedMask(fixed_structure)

    if isinstance(reg_method, str):
        if reg_method.lower() == "translation":
            registration.SetInitialTransform(sitk.TranslationTransform(3))
        elif reg_method.lower() == "similarity":
            registration.SetInitialTransform(sitk.Similarity3DTransform())
        elif reg_method.lower() == "affine":
            registration.SetInitialTransform(sitk.AffineTransform(3))
        elif reg_method.lower() == "rigid":
            registration.SetInitialTransform(sitk.VersorRigid3DTransform())
        elif reg_method.lower() == "scale":
            registration.SetInitialTransform(sitk.ScaleTransform(3))
        elif reg_method.lower() == "scaleversor":
            registration.SetInitialTransform(sitk.ScaleVersor3DTransform())
        elif reg_method.lower() == "scaleskewversor":
            registration.SetInitialTransform(sitk.ScaleSkewVersor3DTransform())
        else:
            raise ValueError(
                "You have selected a registration method that does not exist.\n Please select from"
                " Translation, Similarity, Affine, Rigid, ScaleVersor, ScaleSkewVersor"
            )
    elif isinstance(
        reg_method,
        (
            sitk.CompositeTransform,
            sitk.Transform,
            sitk.TranslationTransform,
            sitk.Similarity3DTransform,
            sitk.AffineTransform,
            sitk.VersorRigid3DTransform,
            sitk.ScaleVersor3DTransform,
            sitk.ScaleSkewVersor3DTransform,
        ),
    ):
        registration.SetInitialTransform(reg_method)
    else:
        raise ValueError(
            "'reg_method' must be either a string (see docs for acceptable registration names), "
            "or a custom sitk.CompositeTransform."
        )

    if optimiser.lower() == "lbfgsb":
        registration.SetOptimizerAsLBFGSB(
            gradientConvergenceTolerance=1e-5,
            numberOfIterations=number_of_iterations,
            maximumNumberOfCorrections=50,
            maximumNumberOfFunctionEvaluations=1024,
            costFunctionConvergenceFactor=1e7,
            trace=verbose,
        )
    elif optimiser.lower() == "exhaustive":
        """
        This isn't well implemented
        Needs some work to give options for sampling rates
        Use is not currently recommended
        """
        samples = [10, 10, 10, 10, 10, 10]
        registration.SetOptimizerAsExhaustive(samples)
    elif optimiser.lower() == "gradient_descent_line_search":
        registration.SetOptimizerAsGradientDescentLineSearch(
            learningRate=1.0, numberOfIterations=number_of_iterations
        )
    elif optimiser.lower() == "gradient_descent":
        registration.SetOptimizerAsGradientDescent(
            learningRate=1.0, numberOfIterations=number_of_iterations
        )

    if verbose:
        registration.AddCommand(
            sitk.sitkIterationEvent,
            lambda: registration_command_iteration(registration),
        )

    output_transform = registration.Execute(fixed=fixed_image, moving=moving_image)
    # Combine initial and optimised transform
    combined_transform = sitk.CompositeTransform([initial_transform, output_transform])

    # Try to find default value
    if default_value is None:
        default_value = 0

        # Test if image is CT-like
        if sitk.GetArrayViewFromImage(moving_image).min() <= -1000:
            default_value = -1000

    registered_image = apply_transform(
        input_image=moving_image,
        reference_image=fixed_image,
        transform=combined_transform,
        default_value=default_value,
        interpolator=final_interp,
    )

    registered_image = sitk.Cast(registered_image, moving_image_type)

    return registered_image, combined_transform



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Various useful utility functions for atlas based segmentation algorithms.
"""

import numpy as np
import SimpleITK as sitk


def label_to_roi(label, expansion_mm=[0, 0, 0], return_as_list=False):
    """Generates a region of interest (ROI), defined by a starting index (z,y,x)
    and size (s_z, s_y, s_x). This can be used to crop images/labels.

    Args:
        label (sitk.Image | list): Binary label image/mask/structure to define to ROI
        expansion_mm (list, optional): An optional expansion of the box (in each direciton).
            Defaults to [0, 0, 0].

    Returns:
        tuple: crop_box_size, crop_box_starting_index
    """

    if hasattr(label, "__iter__") and not isinstance(label, sitk.Image):
        reference_label = sum(label) > 0
    else:
        reference_label = label > 0

    image_spacing = np.array(reference_label.GetSpacing())

    label_stats_image_filter = sitk.LabelStatisticsImageFilter()
    label_stats_image_filter.Execute(reference_label, reference_label)
    bounding_box = np.array(label_stats_image_filter.GetBoundingBox(1))

    index = [bounding_box[x * 2] for x in range(3)]
    size = [bounding_box[(x * 2) + 1] - bounding_box[x * 2] + 1 for x in range(3)]

    expansion_mm = np.array(expansion_mm)
    expansion = (expansion_mm / image_spacing).astype(int)

    # Avoid starting outside the image
    crop_box_index = np.max([index - expansion, np.array([0, 0, 0])], axis=0)

    # Avoid ending outside the image
    crop_box_size = np.min(
        [
            np.array(reference_label.GetSize()) - crop_box_index,
            np.array(size) + 2 * expansion,
        ],
        axis=0,
    )

    crop_box_size = [int(i) for i in crop_box_size]
    crop_box_index = [int(i) for i in crop_box_index]

    if return_as_list:
        return crop_box_index + crop_box_size

    return crop_box_size, crop_box_index


def crop_to_roi(image, size, index):
    """Utility function for cropping images"""
    return sitk.RegionOfInterest(image, size=size, index=index)


def crop_to_label_extent(image, label, expansion_mm=0):
    """Crop an image to the 3D extent defined by a binary mask (label).

    Args:
        image (SimpleITK.Image): The image to crop.
        label (SimpleITK.Image): The binary mask (label) defining the region to crop to.
        expansion_mm (float | tuple | list, optional): An optional expansion in physical units.
            Can be defined as a single number or iterable: (axial, coronal, sagittal) expansion.
            Defaults to 0.

    Returns:
        SimpleITK.Image: The cropped image.
    """

    if ~hasattr(expansion_mm, "__iter__"):
        expansion_mm = [
            expansion_mm,
        ] * 3

    cbox_s, cbox_i = label_to_roi(label, expansion_mm=expansion_mm)
    return crop_to_roi(image, cbox_s, cbox_i)



import math

import numpy as np
import SimpleITK as sitk
import matplotlib
import matplotlib.pyplot as plt
from platipy.imaging.label.utils import get_com

from platipy.imaging.visualisation.visualiser import ImageVisualiser

from platipy.imaging.dose.dvh import (
    calculate_dvh_for_labels,
    calculate_d_x,
    calculate_v_x,
    calculate_d_cc_x,
)


def is_color_dark(color):
    """Returns true if the color is dark, false otherwise.

    Args:
        color (list): A list of values corresponging to r,g,b

    Returns:
        bool: True if dark, False otherwise
    """

    [r, g, b] = color
    hsp = math.sqrt(0.299 * (r * r) + 0.587 * (g * g) + 0.114 * (b * b))
    if hsp > 0.5:
        return False
    else:
        return True


def roundup(x, nearest):
    """Round up to the nearest multiple.

    Args:
        x (float|int): Value to round up
        nearest (float|int): Multiple to round up to

    Returns:
        int: Rounded up value
    """
    return int(math.ceil(x / float(nearest))) * nearest


def visualise_dose(
    img,
    dose,
    structures,
    dvh=None,
    d_points=None,
    v_points=None,
    d_cc_points=None,
    structure_for_com=None,
    structure_for_limits=None,
    expansion_for_limits=10,
    title="",
    contour_cmap=matplotlib.colormaps.get_cmap("rainbow"),
    dose_cmap=matplotlib.colormaps.get_cmap("inferno"),
    structure_name_dict=None,
    img_vis_kw=None,
):
    """A function to visualise an image with dose grid overlaid along with the DVH and dose
    metrics.

    Args:
        img (sitk.Image): Image on which to overlay the dose grid
        dose (sitk.Image): Dose grid
        structures (dict): Dictionary of structures (corresponding to those provided in DVH, if
            provided) with structure name as key and sitk.Image of mask as value.
        dvh (pd.DataFrame, optional): Dose Volume Histogram, if not provided it will be computed.
            Defaults to None.
        d_points (list|float|int, optional): The points at which to calculate D metrics. Defaults
            to None.
        v_points (list|float|int, optional): The points at which to calculate V metrics. Defaults
            to None.
        d_cc_points (list|float|int, optional): Points at which to calculate Dcc metrics. Defaults
            to None.
        structure_for_com (sitk.Image, optional): Mask to use to compute the centre of mass to cut
            the image slices. Defaults to None.
        structure_for_limits (sitk.Image, optional): Mask to set the limits of the visualisation.
            Defaults to None.
        expansion_for_limits (int, optional): Expansion around the structure for limits. Defaults
            to 10.
        title (str, optional): Title to display on visualisation. Defaults to "".
        contour_cmap (plt.cm.colormap, optional): Matplotlib color map to use for contour colors.
            Defaults to matplotlib.colormaps.get_cmap("rainbow").
        dose_cmap (plt.cm.colormap, optional): Matplotlib color map to use for dose colors.
            Defaults to matplotlib.colormaps.get_cmap("inferno").
        structure_name_dict (dict, optional): Dictionary to map alternative structure names.
            Defaults to None.
        img_vis_kw (dict, optional): Dictionary of keyword arguments to pass to ImageVisualiser.
            Defaults to None.

    Returns:
        matplotlib.figure: The figure, can be saved as usual (fig.savefig(...)).
        pd.DataFrame: The dose metrics computed for display in the table.
    """

    if img_vis_kw is None:
        img_vis_kw = {}

    if dvh is None:
        dvh = calculate_dvh_for_labels(dose, structures)
    else:
        dvh = dvh.copy()

    df_metrics = dvh[["label", "mean"]]

    if d_points is not None:
        df_metrics_d = calculate_d_x(dvh, d_points)
        df_metrics = df_metrics.merge(df_metrics_d)

    if v_points is not None:
        df_metrics_v = calculate_v_x(dvh, v_points)
        df_metrics = df_metrics.merge(df_metrics_v)

    if d_cc_points is not None:
        df_metrics_d_cc = calculate_d_cc_x(dvh, d_cc_points)
        df_metrics = df_metrics.merge(df_metrics_d_cc)

    df_metrics = df_metrics.set_index("label")

    cut = None

    if structure_for_com is None:
        # Not structure specified, get the area of highest dose and get the COM of that
        cut = get_com(dose > dose * 0.9)
    else:
        # the user has specified a structure
        cut = get_com(structures[structure_for_com])

    img_vis_kw["cut"] = cut

    vis = ImageVisualiser(img, **img_vis_kw)

    # Resample the dose grid into the image space
    dose = sitk.Resample(dose, img)

    # Cut off lowest 10% of dose for visualisation
    arr = sitk.GetArrayFromImage(dose)
    arr[arr < arr.max() * 0.1] = 0
    dose = sitk.GetImageFromArray(arr)
    dose.CopyInformation(img)

    round_to_nearest = 5
    if arr.max() < 20:
        round_to_nearest = 1

    if arr.max() < 2:
        round_to_nearest = 0.1

    max_val = roundup(arr.max(), round_to_nearest)

    vis.add_scalar_overlay(
        dose,
        discrete_levels=int(max_val / round_to_nearest),
        colormap=dose_cmap,
        alpha=0.5,
        max_value=max_val,
        name="Dose (Gy)",
    )
    vis.add_contour(structures, show_legend=False, colormap=contour_cmap)

    if structure_for_limits is not None:
        vis.set_limits_from_label(structure_for_limits, expansion=expansion_for_limits)

    fig = vis.show()

    dvh.index = dvh.label
    dvh.transpose()
    plt_dvh = dvh.iloc[:, 3:].transpose()

    ax = fig.axes[1]

    subax_x = (ax.bbox.x0 / fig.bbox.width) + 0.175
    subax_y = (ax.bbox.y0 / fig.bbox.height) + 0.05

    subax = fig.add_axes([subax_x, subax_y, 1 - subax_x, 1 - subax_y - 0.3])
    plt_dvh.plot(ax=subax, colormap=contour_cmap, legend=False)

    plt.xlabel("Dose (Gy)")
    plt.ylabel("Frequency")
    plt.title("Dose Volume Histogram (DVH)")

    if structure_name_dict is not None:
        rows = [structure_name_dict[i] for i in list(df_metrics.index)]
    else:
        rows = list(df_metrics.index)
    columns = list(df_metrics.columns)
    cell_text = []
    for _, row in df_metrics.iterrows():
        cell_text.append([f"{s:.2f}" for s in row.values])

    colors = list(matplotlib.colormaps.get_cmap(contour_cmap)(np.linspace(0, 1, len(rows))))

    # Create the table
    table = ax.table(
        cellText=cell_text,
        rowLabels=rows,
        rowColours=colors,
        colLabels=columns,
        fontsize=10,
        bbox=[0.4, 0.5, 0.6, 0.4],
    )

    # Some nice formatting
    for cell in table.get_celld():
        table[cell].set_text_props(va="center")
        table[cell].set_edgecolor("w")

        if cell[0] == 0:  # header
            table[cell].set_text_props(weight="bold", color="w")
            table[cell].set_facecolor("k")

    # Geometry fixes
    for row in range(len(rows) + 1):
        for c in range(len(columns)):
            table[row, c].set_width(0.1)
        if row > 0:
            table[row, -1].set_width(0)

            if is_color_dark(colors[row - 1][:3]):
                table[row, -1].set_text_props(color="w")
            else:
                table[row, -1].set_text_props(color="k")

    for col in range(len(columns)):
        table[0, col].set_facecolor("k")

    table.auto_set_font_size(True)
    font_size = table.get_celld()[1, 0].get_fontsize()

    # insert metadata information
    ax.text(
        x=0.25,
        y=0.96,
        s=title,
        color="black",
        ha="left",
        va="top",
        size=font_size,
        wrap=True,
        weight="bold",
    )

    # Return the figure
    return fig, df_metrics



# Copyright 2020 CSIRO, University of New South Wales, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile

import logging
logger = logging.getLogger(__name__)

from platipy.imaging.projects.bronchus.bronchus import (
    generate_lung_mask,
    generate_airway_mask,
    default_settings
)

BRONCHUS_SETTINGS_DEFAULTS = {
    "outputBronchusName": "Auto_Bronchus",
    "outputLungName": "Auto_Lung",
    "algorithmSettings": default_settings,
}


def run_bronchus_segmentation(input_image, settings=BRONCHUS_SETTINGS_DEFAULTS):
    """Runs the Proximal Bronchial Tree segmentation

    Args:
        input_image (sitk.Image): SimpleITK image on which to perform the segmentation
        settings (dict, optional): Dictionary containing settings for algorithm.
                                   Defaults to BRONCHUS_SETTINGS_DEFAULTS.

    Returns:
        dict: Dictionary containing output of segmentation
    """

    working_directory = tempfile.mkdtemp()
    results = {}

    # Compute the lung mask
    lung_mask = generate_lung_mask(input_image)
    results[settings["outputLungName"]] = lung_mask

    bronchus_mask = generate_airway_mask(
        working_directory,
        input_image,
        lung_mask,
        config_dict=settings["algorithmSettings"],
    )

    # If the bronchus mask counldn't be generated then skip it
    if not bronchus_mask:
        logger.error("Unable to generate bronchus mask")
        return results

    results[settings["outputBronchusName"]] = bronchus_mask

    return results



import SimpleITK as sitk
import numpy as np

from platipy.imaging.label.comparison import compute_surface_dsc, compute_surface_metrics

def test_surface_dsc():

    label_a = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_a.SetSpacing((1,1,2))
    label_a[30:70, 30:70, 30:70] = 1

    # Check contour within acceptable variation
    label_b = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_b.SetSpacing((1,1,2))
    label_b[30:71, 30:71, 30:71] = 1
    sd = compute_surface_dsc(label_a, label_b)
    assert sd == 1.0

    # Move parts of contour away
    label_b = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_b.SetSpacing((1,1,2))
    label_b[35:71, 35:71, 35:71] = 1
    sd = compute_surface_dsc(label_a, label_b)
    assert np.allclose(sd, 0.5158373786407767)

    # Move parts of contour away to shift one dimension out of boundary
    label_b = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_b.SetSpacing((1,1,2))
    label_b[35:72, 35:72, 35:72] = 1
    sd = compute_surface_dsc(label_a, label_b)
    assert np.allclose(sd, 0.39725541227966404)
    
    # Check other dimensions
    label_b = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_b.SetSpacing((1,1,2))
    label_b[35:75, 35:75, 35:75] = 1
    sd = compute_surface_dsc(label_a, label_b)
    assert np.allclose(sd, 0.1258764241893076)

def test_surface_metrics():

    label_a = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_a.SetSpacing((1,1,2))
    label_a[30:70, 30:70, 30:70] = 1

    # Check contour with small shift
    label_b = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_b.SetSpacing((1,1,2))
    label_b[30:71, 30:71, 30:71] = 1
    metrics = compute_surface_metrics(label_a, label_b)
    assert np.allclose(metrics['hausdorffDistance'], 2.449489742783178)
    assert np.allclose(metrics['meanSurfaceDistance'], 0.6649174304423457)
    assert np.allclose(metrics['medianSurfaceDistance'], 0.574099183082580)
    assert np.allclose(metrics['maximumSurfaceDistance'], 2.4494898319244385)
    assert np.allclose(metrics['sigmaSurfaceDistance'], 101.78549149738755)
    assert np.allclose(metrics['surfaceDSC'], 1.0)

    # Check with larger shift
    label_b = sitk.Image(100,100,100, sitk.sitkUInt8)
    label_b.SetSpacing((1,1,2))
    label_b[35:71, 35:71, 35:71] = 1
    metrics = compute_surface_metrics(label_a, label_b)

    assert np.allclose(metrics['hausdorffDistance'], 12.24744871391589)
    assert np.allclose(metrics['meanSurfaceDistance'], 3.842314521867095)
    assert np.allclose(metrics['medianSurfaceDistance'], 3.5163573920726776)
    assert np.allclose(metrics['maximumSurfaceDistance'], 12.24744871391589)
    assert np.allclose(metrics['sigmaSurfaceDistance'], 392.57229390698296)
    assert np.allclose(metrics['surfaceDSC'], 0.5158373786407767)



# Copyright 2020 University of New South Wales, University of Sydney, Ingham Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from skimage.morphology import convex_hull_image

import numpy as np
import SimpleITK as sitk


def get_bone_mask(image, lower_threshold=350, upper_threshold=3500, max_hole_size=5):
    """
    Automatically generate a binary mask of bones from a CT image.

    Args:
        image ([SimpleITK.Image]): The patient x-ray CT image to segment.
        lower_threshold (int, optional): Lower voxel value for threshold. Defaults to 350.
        upper_threshold (int, optional): Upper voxel value for threshold. Defaults to 3500.
        max_hole_size (int | list | bool, optional): Maximum hole size to be filled in millimetres.
                                                     Can be specified as a vector (z,y,x). Defaults
                                                     to 5.

    Returns:
        [SimpleITK.Image]: The binary bone mask.
    """

    bone_mask = sitk.BinaryThreshold(
        image, lowerThreshold=lower_threshold, upperThreshold=upper_threshold
    )

    if max_hole_size is not False:
        if not hasattr(max_hole_size, "__iter__"):
            max_hole_size = (max_hole_size,) * 3

    bone_mask = sitk.BinaryMorphologicalClosing(bone_mask, max_hole_size)

    return bone_mask


def get_external_mask(
    image, lower_threshold=-100, upper_threshold=2500, dilate=1, max_hole_size=False
):
    """
    Automatically generate a binary mask of the patient external contour.
    Uses slice-wise convex hull generation.

    Args:
        image ([SimpleITK.Image]): The patient x-ray CT image to segment. May work with other
                                   modalities with modified thresholds.
        lower_threshold (int, optional): Lower voxel value for threshold. Defaults to -100.
        upper_threshold (int, optional): Upper voxel value for threshold. Defaults to 2500.
        dilate (int | list | bool, optional): Dilation filter size applied to the binary mask. Can
                                              be specified as a vector (z,y,x). Defaults to 1.
        max_hole_size (int  | list | bool, optional): Maximum hole size to be filled in
                                                      millimetres. Can be specified as a vector
                                                      (z,y,x). Defaults to False.

    Returns:
        [SimpleITK.Image]: The binary external mask.
    """

    # Get all points inside the body
    external_mask = sitk.BinaryThreshold(
        image, lowerThreshold=lower_threshold, upperThreshold=upper_threshold
    )

    external_mask_components = sitk.ConnectedComponent(external_mask, True)

    # Second largest volume is most likely the body - you should check this!
    body_mask = sitk.Equal(sitk.RelabelComponent(external_mask_components), 1)

    if dilate is not False:
        if not hasattr(dilate, "__iter__"):
            dilate = (dilate,) * 3
        body_mask = sitk.BinaryDilate(body_mask, dilate)

    if max_hole_size is not False:
        if not hasattr(max_hole_size, "__iter__"):
            max_hole_size = (max_hole_size,) * 3

        body_mask = sitk.BinaryMorphologicalClosing(body_mask, max_hole_size)
        body_mask = sitk.BinaryFillhole(body_mask, fullyConnected=True)

    arr = sitk.GetArrayFromImage(body_mask)

    convex_hull_slices = np.zeros_like(arr)

    for index in np.arange(0, np.alen(arr)):
        convex_hull_slices[index] = convex_hull_image(arr[index])

    body_mask_hull = sitk.GetImageFromArray(convex_hull_slices)
    body_mask_hull.CopyInformation(body_mask)

    return body_mask_hull


def extend_mask(mask, direction=("ax", "sup"), extension_mm=10, interior_mm_shape=10):
    """
    Extends a binary label (mask) a number of slices.
    PROTOTYPE!
    Currently can only extend in axial directions (superior of inferior).
    The shape of the extended part is based on some number of interior slices, as defined.

    Args:
        mask (SimpleITK.Image): The input binary label (mask).
        direction (tuple, optional): The direction as a tuple. First element is axis, second
            element is direction. Defaults to ("ax", "sup").
        extension_mm (int, optional): The extension in millimeters. Defaults to 10.
        interior_mm_shape (int, optional): The length on which to base the extension shape.
            Defaults to 10.

    Returns:
        SimpleITK.Image: The output (extended mask).
    """
    arr = sitk.GetArrayViewFromImage(mask)
    vals = np.unique(arr[arr > 0])
    if len(vals) > 2:
        # There is more than one value! We need to threshold (at the median)
        cutoff = np.median(vals)
        mask_binary = sitk.BinaryThreshold(mask, cutoff, np.max(vals).astype(float))
    else:
        mask_binary = mask

    arr = sitk.GetArrayFromImage(mask_binary)

    if direction[0] == "ax":
        inferior_slice = np.where(arr)[0].min()
        superior_slice = np.where(arr)[0].max()

        n_slices_ext = int(extension_mm / mask.GetSpacing()[2])
        n_slices_est = int(interior_mm_shape / mask.GetSpacing()[2])

        if direction[1] == "sup":
            max_index = min([arr.shape[0], superior_slice + 1 + n_slices_ext])
            for s_in in range(superior_slice + 1 - n_slices_est, max_index):
                arr[s_in, :, :] = np.max(
                    arr[superior_slice - n_slices_est : superior_slice, :, :], axis=0
                )
        if direction[1] == "inf":
            min_index = max([arr.shape[0], inferior_slice - n_slices_ext + n_slices_est])
            for s_in in range(min_index, inferior_slice):
                arr[s_in, :, :] = np.max(
                    arr[inferior_slice + n_slices_est : inferior_slice, :, :], axis=0
                )

    mask_ext = sitk.GetImageFromArray(arr)
    mask_ext.CopyInformation(mask)

    return mask_ext
